from __future__ import annotations

import re
from collections.abc import Mapping

from pydantic import (
    Field,
    computed_field,
    field_validator,
)

from ophyd_async.core import (
    ConfinedModel,
    Device,
    DeviceConnector,
    DeviceFiller,
    DeviceVector,
    LazyMock,
    Signal,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
    gather_dict,
)

from ._epics_connector import fill_backend_with_prefix
from ._signal import PvaSignalBackend, pvget_with_timeout


class PviDeviceConnector(DeviceConnector):
    """Connect to PVI structure served over PVA.

    At init, fill in all the type hinted signals. At connection check their
    types and fill in any extra signals.

    :param prefix:
        The PV prefix of the device, "PVI" will be appended to it to get the PVI
        PV.
    :param error_hint:
        If given, this will be appended to the error message if any of they type
        hinted Signals are not present.
    """

    mock_device_vector_len: int = 2
    pvi_tree: PviTree | None = None

    def __init__(self, prefix: str = "", error_hint: str = "") -> None:
        # TODO: what happens if we get a leading "pva://" here?
        self.prefix = prefix
        self.pvi_pv = prefix + "PVI"
        self.error_hint = error_hint

    def create_children_from_annotations(self, device: Device):
        if not hasattr(self, "filler"):
            self.filler = DeviceFiller(
                device=device,
                signal_backend_factory=PvaSignalBackend,
                device_connector_factory=PviDeviceConnector,
            )
            # Devices will be created with unfilled PviDeviceConnectors
            list(self.filler.create_devices_from_annotations(filled=False))
            # Signals can be filled in with EpicsSignalSuffix and checked at runtime
            for backend, annotations in self.filler.create_signals_from_annotations(
                filled=False
            ):
                fill_backend_with_prefix(self.prefix, backend, annotations)
            self.filler.check_created()

    async def connect_mock(self, device: Device, mock: LazyMock):
        if isinstance(device, DeviceVector):
            self.filler.create_device_vector_entries_to_mock(
                self.mock_device_vector_len
            )
        # Set the name of the device to name all children
        device.set_name(device.name)
        return await super().connect_mock(device, mock)

    async def connect_real(
        self, device: Device, timeout: float, force_reconnect: bool
    ) -> None:
        if not self.pvi_tree:
            # Top-level device, so discover PVI tree
            self.pvi_tree = await PviTree.build_device_tree(
                pvi_pv=self.pvi_pv, timeout=timeout
            )

        # Fill all sub devices
        for device_name, device_sub_tree in self.pvi_tree.sub_devices.items():
            if device_sub_tree.vector_children:
                # This is a DeviceVector
                connector = self.filler.fill_child_device(
                    device_name, device_type=DeviceVector
                )
            else:
                # This is a Device
                connector = self.filler.fill_child_device(device_name)
            connector.pvi_tree = device_sub_tree
            connector.pvi_pv = device_sub_tree.pvi_pv

        # Fill all vector sub-device
        for (
            vector_index,
            vector_child,
        ) in self.pvi_tree.vector_children.items():
            if self.pvi_tree.is_signal_vector:
                # DeviceVector of signals
                if isinstance(vector_child, SignalDetails):
                    backend = self.filler.fill_child_signal(
                        device.name, vector_child.signal_type, vector_index
                    )
                    backend.read_pv = vector_child.read_pv
                    backend.write_pv = vector_child.write_pv
                else:
                    raise TypeError(
                        "Failed to fill DeviceVector. "
                        f"Expected SignalDetails, got {type(vector_child)}"
                    )
            else:
                # DeviceVector of devices
                if isinstance(vector_child, PviTree):
                    connector = self.filler.fill_child_device(
                        device.name, vector_index=vector_index
                    )
                    connector.pvi_tree = vector_child
                    connector.pvi_pv = vector_child.pvi_pv
                else:
                    raise TypeError(
                        "Failed to fill DeviceVector. "
                        f"Expected PviTree, got {type(vector_child)}"
                    )

        # Fill all signals
        for signal_name, signal_details in self.pvi_tree.signals.items():
            backend = self.filler.fill_child_signal(
                signal_name, signal_details.signal_type, None
            )
            backend.read_pv = signal_details.read_pv
            backend.write_pv = signal_details.write_pv

        # Check that all the requested children have been filled
        suffix = f"\n{self.error_hint}" if self.error_hint else ""
        self.filler.check_filled(f"{self.pvi_pv}: {self.pvi_tree}{suffix}")
        # Set the name of the device to name all children
        device.set_name(device.name)
        return await super().connect_real(device, timeout, force_reconnect)


class SignalDetails(ConfinedModel):
    """Representation of a Signal to be constructed."""

    signal_type: type[Signal]
    read_pv: str
    write_pv: str

    @classmethod
    def from_entry(cls, entry: dict[str, str]) -> SignalDetails:
        match entry:
            case {"r": read_pv, "w": write_pv}:
                return cls(signal_type=SignalRW, read_pv=read_pv, write_pv=write_pv)

            case {"rw": pv}:
                return cls(signal_type=SignalRW, read_pv=pv, write_pv=pv)

            case {"r": read_pv}:
                return cls(signal_type=SignalR, read_pv=read_pv, write_pv=read_pv)

            case {"w": write_pv}:
                return cls(signal_type=SignalW, read_pv=write_pv, write_pv=write_pv)

            case {"x": execute_pv}:
                return cls(signal_type=SignalX, read_pv=execute_pv, write_pv=execute_pv)

            case _:
                raise TypeError(f"Can't process entry {entry}")


class PviTree(ConfinedModel):
    """Representation of a PVI structure of devices and signals in a PVI query.

    Example 1: A device with sub-devices and signals
    --------------------------------------
    For a PVI structure such as:

    ```json
    {
        "bit": {"d": "TEST-PANDA:Bits:PVI"},
        "calc": {"d": "TEST-PANDA:Calc:PVI"},
        "a": {"rw": "TEST-PANDA:Bits:A"}
    }
    ```

    From "TEST-PANDA:PVI", This would be represented as:

    ```python
    PviTree(
        pvi_pv="TEST-PANDA:PVI",
        signals={
            "a": SignalDetails(
                signal_type=SignalRW,
                read_pv="TEST-PANDA:Bits:A",
                write_pv="TEST-PANDA:Bits:A")
        },
        sub_devices={
            "bit": PviTree(...),
            "calc": PviTree(...)
        },
        vector_children=[]
    )
    ```

    Example 2: A device with vector children
    -----------------------------------------
    If an entry like `"calc"` is a **DeviceVector**
    (e.g., mirroring a fastCS controller vector), the PVI entries will look like this:

    ```json
    {
        "__1": {"d": "TEST-PANDA:Calc:2:PVI"},
        "__2": {"d": "TEST-PANDA:Calc:1:PVI"}
    }
    ```

    This would be represented as:

    ```python
    PviTree(
        pvi_pv="TEST-PANDA:Calc:PVI",
        signals={},
        sub_devices={},
        vector_children=[
            PviTree(pvi_pv="TEST-PANDA:Calc:2:PVI", signals={}, ...),
            PviTree(pvi_pv="TEST-PANDA:Calc:1:PVI", signals={}, ...)
        ]
    )
    ```

    This is similar for vectors of signals, where `vector_children` would instead
    be populated with `SignalDetails`

    Example 3: A device with legacy vector children
    -----------------------------------------
    Legacy PVI vector structure is supported, for backwards compatability
    with pandablocks-ioc, where vector children are represented as:

    ```
    {
        "calc": [None, {"d": "TEST-PANDA:Calc1:PVI"}, {"d": "TEST-PANDA:Calc2:PVI"}],
    }
    ```
    generate the same PviTree as in Example 2, excluding a PVI PV.

    :param pvi_pv:
        The PVI PV of the device.

    :param signals:
        A mapping of signal names to `SignalDetails` objects.

    :param sub_devices:
        A mapping of sub-device names to their corresponding `PviTree` objects.

    :param vector_children:
        A mapping of int to `PviTree` objects representing child devices of a vector
        device.

    :attr is_signal_vector:
        A computed property returning True if any child device in `vector_children`
        is an instance of `SignalDetails`, else False.
    """

    pvi_pv: str = Field(default="")
    signals: Mapping[str, SignalDetails] = Field(default_factory=dict)
    sub_devices: Mapping[str, PviTree] = Field(default_factory=dict)
    vector_children: Mapping[int, PviTree | SignalDetails] = Field(default_factory=dict)

    @classmethod
    async def build_device_tree(cls, pvi_pv: str, timeout: float) -> PviTree:
        """Recursively build a PviTree from a top level device.

        Starting from the top-level device, this classmethod performs
        post-order traversal over the served PVI structure, populating
        a PviTree from the bottom up.

        :param name: Device name
        :param pvi_pv: Device PVI PV
        :param timeout: Timeout on pvget
        """
        pvi_structure = await pvget_with_timeout(pvi_pv, timeout)

        # An example entry is: {"d": "Prefix:Device:PVI", "rw": "Prefix:A"}
        # these entries are stored under the parent PVI structure name
        # for example, {"device": {"d": "Prefix:Device:PVI", "rw": "Prefix:A"}}
        entries: dict[str, dict[str, str]] = pvi_structure["value"].todict()
        signal_details = {
            entry_name: SignalDetails.from_entry(entries.pop(entry_name))
            for entry_name in list(entries)
            if not isinstance(entries[entry_name], list)
            and set(entries[entry_name]) != {"d"}
        }

        sub_trees = await gather_dict(
            {
                entry_name: cls._handle_legacy_entry(entry, timeout)
                if isinstance(entry, list)  # Found a legacy entry, try to handle
                else cls.build_device_tree(entry["d"], timeout)
                for entry_name, entry in entries.items()
            }
        )

        vector_children: dict[int, PviTree | SignalDetails] = {}
        # Filter vector children out of stand-alone devices

        for processed_entries in (sub_trees, signal_details):
            for child_name in list(processed_entries):
                if m := re.match(r"^__(\d+)$", child_name):
                    sub_tree = processed_entries.pop(child_name)
                    vector_children[int(m.group(1))] = sub_tree

        return PviTree(
            pvi_pv=pvi_pv,
            signals=signal_details,
            sub_devices=sub_trees,
            vector_children=vector_children,
        )

    @classmethod
    async def _handle_legacy_entry(
        cls, legacy_entry: list[None | dict[str, str]], timeout: float
    ) -> PviTree:
        """Handle legacy vector entries.

        For example;
        ```
        {
            "calc": [None, {"d": "TEST-PANDA:Calc1:PVI"}, {"d": "TEST-PANDA:Calc2:PVI"}]
        }
        ```

        a `PviTree` is built for each device entry in this list.
        """
        sub_trees = await gather_dict(
            {
                vector_index: cls.build_device_tree(vector_entry["d"], timeout)
                for vector_index, vector_entry in enumerate(legacy_entry)
                if vector_entry is not None
            }
        )

        # Legacy FastCS vector should not contain child signals,
        # devices, or its own PVI PV.
        return PviTree(
            vector_children=sub_trees,
        )

    @computed_field
    @property
    def is_signal_vector(self) -> bool:
        """Flags if a PviTree represents a DeviceVector of Signals or Devices."""
        return any(isinstance(v, SignalDetails) for v in self.vector_children.values())

    def __str__(self) -> str:
        """Print a readable top layer of the PviTree."""
        children = {
            **{
                child_name: tree.pvi_pv for child_name, tree in self.sub_devices.items()
            },
            **{
                child_name: vector_child.pvi_pv
                for child_name, vector_child in self.vector_children.items()
                if isinstance(vector_child, PviTree)
            },
        }

        signals = {
            **{
                signal_name: (
                    detail.signal_type.__name__,
                    detail.read_pv,
                    detail.write_pv,
                )
                for signal_name, detail in self.signals.items()
            },
            **{
                signal_name: (
                    vector_child.signal_type.__name__,
                    vector_child.read_pv,
                    vector_child.write_pv,
                )
                for signal_name, vector_child in self.vector_children.items()
                if isinstance(vector_child, SignalDetails)
            },
        }
        return f"sub_devices={children}\nsignals={signals}"

    @field_validator("vector_children")
    @classmethod
    def _check_consistency_of_vector_children_type(
        cls,
        vector_children: Mapping[int, PviTree | SignalDetails],
    ):
        """Validates that parsed vector children are all of the same type."""
        if not (
            all(
                isinstance(vector_child, SignalDetails)
                for vector_child in vector_children.values()
            )
            or all(
                isinstance(vector_child, PviTree)
                for vector_child in vector_children.values()
            )
        ):
            raise ValueError(
                "Failed to validate PviTree. "
                "vector_children must all be of type `SignalDetails` or `PviTree`. "
                f"Received mixed type: {vector_children=}"
            )
        return vector_children



from ophyd_async.core import Device, SignalR, SignalRW, SignalX


class MetaWriterIO(Device):
    """Ophyd-async implementation of a MetaWriter Odin Subdevice."""

    stop: SignalX
    file_prefix: SignalRW[str]
    directory: SignalRW[str]
    acquisition_id: SignalRW[str]
    writing: SignalR[bool]


class FrameProcessorIO(Device):
    """Ophyd-async implementation of a FrameProcessor Odin Subdevice."""

    start_writing: SignalX
    stop_writing: SignalX
    writing: SignalR[bool]
    frames_written: SignalR[int]
    frames: SignalRW[int]
    data_dims_0: SignalRW[int]
    data_dims_1: SignalRW[int]
    data_chunks_0: SignalRW[int]
    data_chunks_1: SignalRW[int]
    data_chunks_2: SignalRW[int]
    file_path: SignalRW[str]
    file_prefix: SignalRW[str]
    data_datatype: SignalRW[str]
    data_compression: SignalRW[str]
    process_frames_per_block: SignalRW[int]


class OdinIO(Device):
    fp: FrameProcessorIO
    mw: MetaWriterIO



"""Support for the ADVimba areaDetector driver.

https://github.com/areaDetector/ADVimba.
"""

import asyncio
from collections.abc import Sequence
from typing import Annotated as A

from ophyd_async.core import (
    DetectorTriggerLogic,
    OnOff,
    PathProvider,
    SignalDict,
    SignalR,
    SignalRW,
    StrictEnum,
)
from ophyd_async.epics.core import PvSuffix

from .adcore import (
    ADArmLogic,
    ADBaseIO,
    ADWriterType,
    AreaDetector,
    NDPluginBaseIO,
    prepare_exposures,
)
from .adgenicam import get_camera_deadtime

__all__ = [
    "VimbaDetector",
    "VimbaDriverIO",
    "VimbaTriggerLogic",
    "VimbaConvertFormat",
    "VimbaTriggerSource",
    "VimbaOverlap",
    "VimbaExposeOutMode",
]


class VimbaConvertFormat(StrictEnum):
    """Convert pixel format for the Vimba detector."""

    NONE = "None"
    MONO8 = "Mono8"
    MONO16 = "Mono16"
    RGB8 = "RGB8"
    RGB16 = "RGB16"


class VimbaTriggerSource(StrictEnum):
    """Mode for the source of triggers on the Vimba."""

    FREERUN = "Freerun"
    LINE1 = "Line1"
    LINE2 = "Line2"
    FIXED_RATE = "FixedRate"
    SOFTWARE = "Software"
    ACTION0 = "Action0"
    ACTION1 = "Action1"


class VimbaOverlap(StrictEnum):
    """Overlap modes for the Vimba detector."""

    OFF = OnOff.OFF.value
    PREV_FRAME = "PreviousFrame"


class VimbaExposeOutMode(StrictEnum):
    """Exposure control modes for Vimba detectors."""

    TIMED = "Timed"  # Use ExposureTime PV
    TRIGGER_WIDTH = "TriggerWidth"  # Expose for length of high signal


class VimbaDriverIO(ADBaseIO):
    """Mirrors the interface provided by ADVimba/db/vimba.template."""

    convert_pixel_format: A[
        SignalRW[VimbaConvertFormat], PvSuffix.rbv("ConvertPixelFormat")
    ]
    trigger_source: A[SignalRW[VimbaTriggerSource], PvSuffix.rbv("TriggerSource")]
    trigger_mode: A[SignalRW[OnOff], PvSuffix.rbv("TriggerMode")]
    trigger_overlap: A[SignalRW[VimbaOverlap], PvSuffix.rbv("TriggerOverlap")]
    exposure_mode: A[SignalRW[VimbaExposeOutMode], PvSuffix.rbv("ExposureMode")]


class VimbaTriggerLogic(DetectorTriggerLogic):
    """Trigger logic for ADVimba detectors."""

    def __init__(self, driver: VimbaDriverIO, override_deadtime: float | None = None):
        self.driver = driver
        self.override_deadtime = override_deadtime

    def config_sigs(self) -> set[SignalR]:
        return {self.driver.model}

    def get_deadtime(self, config_values: SignalDict) -> float:
        return get_camera_deadtime(
            model=config_values[self.driver.model],
            override_deadtime=self.override_deadtime,
        )

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        await asyncio.gather(
            self.driver.trigger_mode.set(OnOff.OFF),
            self.driver.exposure_mode.set(VimbaExposeOutMode.TIMED),
            self.driver.trigger_source.set(VimbaTriggerSource.FREERUN),
        )
        await prepare_exposures(self.driver, num, livetime, deadtime)

    async def prepare_edge(self, num: int, livetime: float):
        await asyncio.gather(
            self.driver.trigger_mode.set(OnOff.ON),
            self.driver.exposure_mode.set(VimbaExposeOutMode.TIMED),
            self.driver.trigger_source.set(VimbaTriggerSource.LINE1),
        )
        await prepare_exposures(self.driver, num, livetime)

    async def prepare_level(self, num: int):
        await asyncio.gather(
            self.driver.trigger_mode.set(OnOff.ON),
            self.driver.exposure_mode.set(VimbaExposeOutMode.TRIGGER_WIDTH),
            self.driver.trigger_source.set(VimbaTriggerSource.LINE1),
        )
        await prepare_exposures(self.driver, num)


class VimbaDetector(AreaDetector[VimbaDriverIO]):
    """Create an ADVimba AreaDetector instance.

    :param prefix: EPICS PV prefix for the detector
    :param path_provider: Provider for file paths during acquisition
    :param driver_suffix: Suffix for the driver PV, defaults to "cam1:"
    :param override_deadtime:
        If provided, this value is used for deadtime instead of looking up
        based on camera model.
    :param writer_type: Type of file writer (HDF or TIFF)
    :param writer_suffix: Suffix for the writer PV
    :param plugins: Additional areaDetector plugins to include
    :param config_sigs: Additional signals to include in configuration
    :param name: Name for the detector device
    """

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider | None = None,
        driver_suffix="cam1:",
        override_deadtime: float | None = None,
        writer_type: ADWriterType | None = ADWriterType.HDF,
        writer_suffix: str | None = None,
        plugins: dict[str, NDPluginBaseIO] | None = None,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ) -> None:
        driver = VimbaDriverIO(prefix + driver_suffix)
        super().__init__(
            prefix=prefix,
            driver=driver,
            arm_logic=ADArmLogic(driver),
            trigger_logic=VimbaTriggerLogic(driver, override_deadtime),
            path_provider=path_provider,
            writer_type=writer_type,
            writer_suffix=writer_suffix,
            plugins=plugins,
            config_sigs=config_sigs,
            name=name,
        )



"""areaDetector support for Aravis GigE and USB3 vision cameras.

https://github.com/areaDetector/ADAravis
"""

from collections.abc import Sequence
from typing import Annotated as A

from ophyd_async.core import (
    DetectorTriggerLogic,
    OnOff,
    PathProvider,
    SignalDict,
    SignalR,
    SignalRW,
    SubsetEnum,
)

from .adcore import (
    ADArmLogic,
    ADBaseIO,
    ADWriterType,
    AreaDetector,
    NDPluginBaseIO,
    prepare_exposures,
)
from .adgenicam import get_camera_deadtime
from .core import PvSuffix

__all__ = [
    "AravisDetector",
    "AravisDriverIO",
    "AravisTriggerLogic",
    "AravisTriggerSource",
]


class AravisTriggerSource(SubsetEnum):
    """Which trigger source to use when TriggerMode=On."""

    LINE1 = "Line1"


class AravisDriverIO(ADBaseIO):
    """Generic Driver supporting all GiGE cameras.

    This mirrors the interface provided by ADAravis/db/aravisCamera.template.
    """

    trigger_mode: A[SignalRW[OnOff], PvSuffix.rbv("TriggerMode")]
    trigger_source: A[SignalRW[AravisTriggerSource], PvSuffix.rbv("TriggerSource")]


class AravisTriggerLogic(DetectorTriggerLogic):
    """Trigger logic for Aravis GigE and USB3 cameras."""

    def __init__(self, driver: AravisDriverIO, override_deadtime: float | None = None):
        self.driver = driver
        self.override_deadtime = override_deadtime

    def config_sigs(self) -> set[SignalR]:
        return {self.driver.model}

    def get_deadtime(self, config_values: SignalDict) -> float:
        return get_camera_deadtime(
            model=config_values[self.driver.model],
            override_deadtime=self.override_deadtime,
        )

    async def prepare_internal(self, num: int, livetime: float, deadtime: float):
        await self.driver.trigger_mode.set(OnOff.OFF)
        await prepare_exposures(self.driver, num, livetime, deadtime)

    async def prepare_edge(self, num: int, livetime: float):
        # Trigger on the rising edge of Line1
        # trigger mode must be set first and on its own!
        # Hardware race condition in Aravis firmware requires setting trigger mode
        # separately before trigger source to avoid undefined behavior.
        # https://github.com/AravisProject/aravis/issues/1045
        await self.driver.trigger_mode.set(OnOff.ON)
        await self.driver.trigger_source.set(AravisTriggerSource.LINE1)
        await prepare_exposures(self.driver, num, livetime)


class AravisDetector(AreaDetector[AravisDriverIO]):
    """Create an ADAravis AreaDetector instance.

    :param prefix: EPICS PV prefix for the detector
    :param path_provider: Provider for file paths during acquisition
    :param driver_suffix: Suffix for the driver PV, defaults to "cam1:"
    :param override_deadtime:
        If provided, this value is used for deadtime instead of looking up
        based on camera model.
    :param writer_type: Type of file writer (HDF or TIFF)
    :param writer_suffix: Suffix for the writer PV
    :param plugins: Additional areaDetector plugins to include
    :param config_sigs: Additional signals to include in configuration
    :param name: Name for the detector device
    """

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider | None = None,
        driver_suffix="cam1:",
        override_deadtime: float | None = None,
        writer_type: ADWriterType | None = ADWriterType.HDF,
        writer_suffix: str | None = None,
        plugins: dict[str, NDPluginBaseIO] | None = None,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ) -> None:
        driver = AravisDriverIO(prefix + driver_suffix)
        super().__init__(
            prefix=prefix,
            driver=driver,
            arm_logic=ADArmLogic(driver),
            trigger_logic=AravisTriggerLogic(driver, override_deadtime),
            path_provider=path_provider,
            writer_type=writer_type,
            writer_suffix=writer_suffix,
            plugins=plugins,
            config_sigs=config_sigs,
            name=name,
        )



import io
import logging
from unittest.mock import MagicMock, patch

import pytest

from ophyd_async.core import Device, _log, config_ophyd_async_logging  # noqa: PLC2701

# Allow this importing of _log for now to test the internal interface
# But this needs resolving.


def test_validate_level():
    assert _log._validate_level("CRITICAL") == 50
    assert _log._validate_level("ERROR") == 40
    assert _log._validate_level("WARNING") == 30
    assert _log._validate_level("INFO") == 20
    assert _log._validate_level("DEBUG") == 10
    assert _log._validate_level("NOTSET") == 0
    assert _log._validate_level(123) == 123
    with pytest.raises(ValueError):
        _log._validate_level("MYSTERY")


def test_default_config_ophyd_async_logging():
    config_ophyd_async_logging()
    assert isinstance(_log.current_handler, logging.StreamHandler)
    assert _log.logger.getEffectiveLevel() <= logging.WARNING


def test_config_ophyd_async_logging_with_file_handler(tmp_path):
    config_ophyd_async_logging(file=tmp_path / "file")
    assert isinstance(_log.current_handler, logging.StreamHandler)
    assert _log.logger.getEffectiveLevel() <= logging.WARNING


def test_config_ophyd_async_logging_removes_extra_handlers():
    # Protect global variable in other pytests
    class FakeLogger:
        def __init__(self):
            self.handlers = []
            self.removeHandler = MagicMock()
            self.setLevel = MagicMock()

        def addHandler(self, handler):  # noqa: N802
            self.handlers.append(handler)

        def getEffectiveLevel(self):  # noqa: N802
            return 100000

    fake_logger = FakeLogger()
    with (
        patch("ophyd_async.core._log.logger", fake_logger),
    ):
        config_ophyd_async_logging()
        fake_logger.removeHandler.assert_not_called()
        config_ophyd_async_logging()
        fake_logger.removeHandler.assert_called()


# Full format looks like:
# '[test_device][W 240501 13:28:08.937 test_log:35] here is a warning\n'
def test_logger_adapter_ophyd_async_device():
    log_buffer = io.StringIO()
    log_stream = logging.StreamHandler(stream=log_buffer)
    log_stream.setFormatter(
        _log.ColoredFormatterWithDeviceName(
            fmt=_log.DEFAULT_FORMAT, datefmt=_log.DEFAULT_DATE_FORMAT, no_color=True
        )
    )
    _log.logger.addHandler(log_stream)

    device = Device(name="test_device")
    device._log = logging.LoggerAdapter(
        logging.getLogger("ophyd_async.devices"),
        {"ophyd_async_device_name": device.name},
    )
    device._log.warning("here is a warning")
    assert log_buffer.getvalue().startswith("[test_device]")
    assert log_buffer.getvalue().endswith("here is a warning\n")



from abc import abstractmethod
from collections.abc import AsyncIterator, Sequence
from dataclasses import dataclass
from typing import Any

from bluesky.protocols import Reading, StreamAsset
from event_model import ComposeStreamResource, DataKey, StreamRange

from ._signal import SignalR, SignalW
from ._utils import ConfinedModel


class ReadableDataProvider:
    @abstractmethod
    async def make_datakeys(self) -> dict[str, DataKey]:
        """Return a DataKey for each Readable that produces a Reading.

        Called before the first exposure is taken.
        """

    @abstractmethod
    async def make_readings(self) -> dict[str, Reading]:
        """Read the Signals and return their values."""


@dataclass
class SignalDataProvider(ReadableDataProvider):
    signal: SignalR

    async def make_datakeys(self) -> dict[str, DataKey]:
        return await self.signal.describe()

    async def make_readings(self) -> dict[str, Reading]:
        return await self.signal.read(cached=False)


class StreamableDataProvider:
    collections_written_signal: SignalR[int]

    @abstractmethod
    async def make_datakeys(self, collections_per_event: int) -> dict[str, DataKey]:
        """Return a DataKey for each Signal that produces a Reading.

        Called before the first exposure is taken.

        :param collections_per_event: this should appear in the shape of each DataKey
        """

    async def make_stream_docs(
        self, collections_written: int, collections_per_event: int
    ) -> AsyncIterator[StreamAsset]:
        """Make StreamAsset documents up to the given index.

        Default implementation is a no-op. Subclasses should override this
        to emit actual StreamAsset documents.
        """
        while False:
            yield


class StreamResourceInfo(ConfinedModel):
    """A description of a single StreamResource that should be emitted."""

    data_key: str
    """The data_key that will appear in the event descriptor,
    e.g. det or det.data"""

    shape: tuple[int | None, ...]
    """The shape of a single collection's data in the HDF file,
    e.g. (768, 1024) for arrays or () for scalars"""

    chunk_shape: tuple[int, ...]
    """The explicit chunk size written to disk"""

    dtype_numpy: str
    """The numpy dtype for this field,
    e.g. <i2 or <f8"""

    parameters: dict[str, Any]
    """Any other parameters that should be included in the StreamResource,
    e.g. dataset path"""

    source: str = ""
    """The source string that should appear in the event descriptor, blank means use uri
    e.g. ca://HDF:FullFileName_RBV"""


class StreamResourceDataProvider(StreamableDataProvider):
    """A helper class to make stream resource and datums for HDF datasets.

    :param full_file_name: Absolute path to the file that has been written
    :param datasets: Descriptions of each of the datasets that will appear in the file
    """

    def __init__(
        self,
        uri: str,
        resources: Sequence[StreamResourceInfo],
        mimetype: str,
        collections_written_signal: SignalR[int],
        flush_signal: SignalW[bool] | None = None,
    ) -> None:
        self.uri = uri
        self.resources = list(resources)
        self.collections_written_signal = collections_written_signal
        self.flush_signal = flush_signal
        self.last_emitted = 0
        bundler_composer = ComposeStreamResource()
        self.bundles = [
            bundler_composer(
                mimetype=mimetype,
                uri=uri,
                data_key=resource.data_key,
                parameters={
                    "chunk_shape": resource.chunk_shape,
                    **resource.parameters,
                },
                uid=None,
                validate=True,
            )
            for resource in self.resources
        ]

    async def make_datakeys(self, collections_per_event: int) -> dict[str, DataKey]:
        describe = {
            resource.data_key: DataKey(
                source=resource.source or self.uri,
                shape=[collections_per_event, *resource.shape],
                dtype="array"
                if collections_per_event > 1 or len(resource.shape) > 1
                else "number",
                dtype_numpy=resource.dtype_numpy,
                external="STREAM:",
            )
            for resource in self.resources
        }
        return describe

    async def make_stream_docs(
        self, collections_written: int, collections_per_event: int
    ) -> AsyncIterator[StreamAsset]:
        if self.flush_signal:
            await self.flush_signal.set(True)
        # TODO: fail if we get dropped frames
        indices_written = collections_written // collections_per_event
        if indices_written and not self.last_emitted:
            for bundle in self.bundles:
                yield "stream_resource", bundle.stream_resource_doc

        if indices_written > self.last_emitted:
            indices: StreamRange = {
                "start": self.last_emitted,
                "stop": indices_written,
            }
            self.last_emitted = indices_written
            for bundle in self.bundles:
                yield "stream_datum", bundle.compose_stream_datum(indices)



import asyncio
import os
import typing
from collections.abc import Callable, Sequence
from typing import Any, TypeVar

import numpy as np
import pytest
from bluesky.protocols import Reading

from ophyd_async.core import (
    Array1D,
    SoftSignalBackend,
    StrictEnum,
    soft_signal_rw,
)
from ophyd_async.testing import ExampleEnum, ExampleTable, MonitorQueue

T = TypeVar("T")


class MyEnum(StrictEnum):
    A = "Aaa"
    B = "Bbb"
    C = "Ccc"


def integer_d(value):
    return {"dtype": "integer", "shape": []}


def number_d(value):
    return {"dtype": "number", "shape": []}


def string_d(value):
    return {"dtype": "string", "shape": []}


def enum_d(value):
    return {"dtype": "string", "shape": [], "choices": ["Aaa", "Bbb", "Ccc"]}


def waveform_d(value):
    return {"dtype": "array", "shape": [len(value)]}


def enumwf_d(value):
    return {"dtype": "array", "shape": [len(value)], "choices": ["Aaa", "Bbb", "Ccc"]}


def table_d(value):
    return {"dtype": "array", "shape": [len(value)]}


# Can be removed once numpy >=2 is pinned.
scalar_int_dtype = (
    "<i4" if os.name == "nt" and np.version.version.startswith("1.") else "<i8"
)


@pytest.mark.parametrize(
    "datatype, initial_value, put_value, descriptor, dtype_numpy",
    [
        (int, 0, 43, integer_d, scalar_int_dtype),
        (float, 0.0, 43.5, number_d, "<f8"),
        (str, "", "goodbye", string_d, "|S40"),
        (MyEnum, MyEnum.A, MyEnum.C, enum_d, "|S40"),
        (Array1D[np.int8], np.array([]), np.array([-8, 3, 44]), waveform_d, "|i1"),
        (Array1D[np.uint8], np.array([]), np.array([218]), waveform_d, "|u1"),
        (Array1D[np.int16], np.array([]), np.array([-855]), waveform_d, "<i2"),
        (Array1D[np.uint16], np.array([]), np.array([5666]), waveform_d, "<u2"),
        (Array1D[np.int32], np.array([]), np.array([-2]), waveform_d, "<i4"),
        (Array1D[np.uint32], np.array([]), np.array([1022233]), waveform_d, "<u4"),
        (Array1D[np.int64], np.array([]), np.array([-3]), waveform_d, "<i8"),
        (Array1D[np.uint64], np.array([]), np.array([995444]), waveform_d, "<u8"),
        (Array1D[np.float32], np.array([]), np.array([1.0]), waveform_d, "<f4"),
        (Array1D[np.float64], np.array([]), np.array([0.2]), waveform_d, "<f8"),
        (Sequence[str], [], ["nine", "ten"], waveform_d, "|S40"),
        (Sequence[MyEnum], [], [MyEnum.A, MyEnum.B], enumwf_d, "|S40"),
        (typing.Sequence[str], [], ["nine", "ten"], waveform_d, "|S40"),
        (typing.Sequence[MyEnum], [], [MyEnum.A, MyEnum.B], enumwf_d, "|S40"),
        (
            ExampleTable,
            ExampleTable(
                a_bool=np.array([]),
                a_int=np.array([]),
                a_float=np.array([]),
                a_str=[],
                a_enum=[],
            ),
            ExampleTable(
                a_bool=np.array([True]),
                a_int=np.array([525]),
                a_float=np.array([3.14]),
                a_str=["pi"],
                a_enum=[ExampleEnum.A],
            ),
            table_d,
            [
                ("a_bool", "|b1"),
                ("a_int", "<i4"),
                ("a_float", "<f8"),
                ("a_str", "|S40"),
                ("a_enum", "|S40"),
            ],
        ),
    ],
)
async def test_soft_signal_backend_get_put_monitor(
    datatype: type[T],
    initial_value: T,
    put_value: T,
    descriptor: Callable[[Any], dict],
    dtype_numpy: str,
):
    signal = soft_signal_rw(datatype, initial_value)
    await signal.connect()
    backend = signal._connector.backend
    with MonitorQueue(signal) as q:
        # Check descriptor
        source = "soft://test"
        # Add expected dtype_numpy to descriptor
        assert dict(
            source=source, **descriptor(initial_value), dtype_numpy=dtype_numpy
        ) == await backend.get_datakey(source)
        await q.assert_updates(initial_value)

        # Put to new value and check that
        await backend.put(put_value)
        await q.assert_updates(put_value)


async def test_soft_signal_backend_enum_value_equivalence():
    soft_backend = SoftSignalBackend(MyEnum)
    await soft_backend.connect(timeout=1)
    assert (await soft_backend.get_value()) is MyEnum.A
    await soft_backend.put(MyEnum.B)
    assert (await soft_backend.get_value()) is MyEnum.B


async def test_soft_signal_backend_set_callback():
    soft_backend = SoftSignalBackend(Array1D[np.float64])
    updates: asyncio.Queue[Reading] = asyncio.Queue()
    # set a callback, so that the subsequent set will fail
    soft_backend.set_callback(updates.put_nowait)
    assert soft_backend.callback is not None
    with pytest.raises(
        RuntimeError, match="Cannot set a callback when one is already set"
    ):
        soft_backend.set_callback(updates.put_nowait)
    soft_backend.set_callback(None)
    assert soft_backend.callback is None


async def test_soft_signal_backend_with_numpy_typing():
    soft_backend = SoftSignalBackend(Array1D[np.float64])
    await soft_backend.connect(timeout=1)
    await soft_backend.put(np.array([1, 2]))
    array = await soft_backend.get_value()
    assert array.shape == (2,)
    assert array[0] == 1


async def test_soft_signal_descriptor_fails_for_invalid_class():
    class MyClass:
        def __init__(self) -> None:
            pass

    with pytest.raises(TypeError):
        SoftSignalBackend(MyClass)


async def test_soft_signal_descriptor_with_metadata():
    soft_signal = soft_signal_rw(int, 0, units="mm", precision=0)
    await soft_signal.connect()
    datakey = await soft_signal.describe()
    assert datakey[""]["units"] == "mm"
    assert datakey[""]["precision"] == 0

    soft_signal = soft_signal_rw(int, units="")
    await soft_signal.connect()
    datakey = await soft_signal.describe()
    assert datakey[""]["units"] == ""
    assert not hasattr(datakey[""], "precision")


async def test_soft_signal_descriptor_with_no_metadata_not_passed():
    soft_signal = soft_signal_rw(int)
    await soft_signal.connect()
    datakey = await soft_signal.describe()
    assert not hasattr(datakey[""], "units")
    assert not hasattr(datakey[""], "precision")


async def test_soft_signal_coerces_numpy_types():
    soft_signal = soft_signal_rw(float)
    await soft_signal.connect()
    assert await soft_signal.get_value() == 0.0
    assert type(await soft_signal.get_value()) is float
    await soft_signal.set(np.float64(1.1))
    assert await soft_signal.get_value() == 1.1
    assert type(await soft_signal.get_value()) is float
    soft_signal._connector.backend.set_value(np.float64(2.2))
    assert await soft_signal.get_value() == 2.2
    assert type(await soft_signal.get_value()) is float



"""Some simulated devices to be used in tutorials and testing."""

from ._blob_detector import SimBlobDetector
from ._mirror_horizontal import HorizontalMirror, HorizontalMirrorDerived
from ._mirror_vertical import (
    TwoJackDerived,
    TwoJackRaw,
    TwoJackTransform,
    VerticalMirror,
)
from ._motor import SimMotor
from ._pattern_generator import PatternGenerator
from ._point_detector import SimPointDetector
from ._stage import SimStage

__all__ = [
    "SimMotor",
    "SimStage",
    "PatternGenerator",
    "SimPointDetector",
    "SimBlobDetector",
    "VerticalMirror",
    "HorizontalMirror",
    "HorizontalMirrorDerived",
    "TwoJackTransform",
    "TwoJackDerived",
    "TwoJackRaw",
]



from collections.abc import Sequence

from ophyd_async.core import PathProvider, SignalR, StandardDetector

from ._blob_arm_logic import BlobArmLogic
from ._blob_data_logic import BlobDataLogic
from ._blob_trigger_logic import BlobTriggerLogic
from ._pattern_generator import PatternGenerator


class SimBlobDetector(StandardDetector):
    """Simulates a detector and writes Blobs to file."""

    def __init__(
        self,
        path_provider: PathProvider,
        pattern_generator: PatternGenerator | None = None,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ) -> None:
        self.pattern_generator = pattern_generator or PatternGenerator()
        self.add_detector_logics(
            BlobTriggerLogic(pattern_generator=self.pattern_generator),
            BlobArmLogic(pattern_generator=self.pattern_generator),
            BlobDataLogic(
                path_provider=path_provider, pattern_generator=self.pattern_generator
            ),
        )
        self.add_config_signals(*config_sigs)
        super().__init__(name=name)



from collections.abc import Iterable, Iterator
from contextlib import contextmanager
from unittest.mock import AsyncMock, Mock

from ._device import Device, DeviceMock
from ._mock_signal_backend import MockPutCallback, MockSignalBackend
from ._signal import Signal, SignalConnector, SignalR
from ._signal_backend import SignalDatatypeT


def get_mock(device: Device | Signal) -> Mock:
    """Return the mock (which may have child mocks attached) for a Device.

    The device must have been connected in mock mode.
    """
    mock = device._mock  # noqa: SLF001
    if not isinstance(mock, DeviceMock):
        msg = f"Device {device} not connected in mock mode"
        raise RuntimeError(msg)
    return mock()


def _get_mock_signal_backend(signal: Signal) -> MockSignalBackend:
    connector = signal._connector  # noqa: SLF001
    if not isinstance(connector, SignalConnector):
        raise TypeError(f"Expected Signal, got {signal}")
    if not isinstance(connector.backend, MockSignalBackend):
        raise RuntimeError(f"Signal {signal} not connected in mock mode")
    return connector.backend


def set_mock_value(signal: Signal[SignalDatatypeT], value: SignalDatatypeT):
    """Set the value of a signal that is in mock mode."""
    backend = _get_mock_signal_backend(signal)
    backend.set_value(value)


class _SetValuesIterator(Iterator[SignalDatatypeT]):
    # Garbage collected by the time __del__ is called unless we put it as a
    # global attrbute here.
    require_all_consumed: bool = False

    def __init__(
        self,
        signal: SignalR[SignalDatatypeT],
        values: Iterable[SignalDatatypeT],
        require_all_consumed: bool = False,
    ):
        self.signal = signal
        self.values = values
        self.require_all_consumed = require_all_consumed
        self.index = 0
        self.iterator = enumerate(values, start=1)

    def __next__(self) -> SignalDatatypeT:
        # Will propogate StopIteration
        self.index, next_value = next(self.iterator)
        set_mock_value(self.signal, next_value)
        return next_value

    def __del__(self):
        if self.require_all_consumed:
            # Values is cast to a list here because the user has supplied
            # require_all_consumed=True, we can therefore assume they
            # supplied a finite list.
            # In the case of require_all_consumed=False, an infinite
            # iterble is permitted
            values = list(self.values)
            if self.index != len(values):
                # Report the values consumed and the values yet to be
                # consumed
                consumed = values[0 : self.index]
                to_be_consumed = values[self.index :]
                raise AssertionError(
                    f"{self.signal.name}: {consumed} were consumed "
                    f"but {to_be_consumed} were not consumed"
                )


def set_mock_values(
    signal: SignalR[SignalDatatypeT],
    values: Iterable[SignalDatatypeT],
    require_all_consumed: bool = False,
) -> Iterator[SignalDatatypeT]:
    """Set a signal to a sequence of values, optionally repeating.

    :param signal: A signal connected in mock mode.
    :param values:
        An iterable of the values to set the signal to, on each iteration the
        next value will be set.
    :param require_all_consumed:
        If True, an AssertionError will be raised if the iterator is deleted
        before all values have been consumed.

    :example:
    ```python
    for value_set in set_mock_values(signal, range(3)):
        # do something

    cm = set_mock_values(signal, [1, 3, 8], require_all_consumed=True):
    next(cm) # do something
    ```
    """
    return _SetValuesIterator(
        signal,
        values,
        require_all_consumed=require_all_consumed,
    )


@contextmanager
def _unset_side_effect_cm(backend: MockSignalBackend):
    yield
    backend.set_mock_put_callback(None)


def callback_on_mock_put(signal: Signal[SignalDatatypeT], callback: MockPutCallback):
    """For setting a callback when a backend is put to.

    Can either be used in a context, with the callback being unset on exit, or
    as an ordinary function.

    The value that the callback returns (if not None) will be set to the signal
    readback. If None is returned then the readback will be set to the setpoint.

    :param signal: A signal with a `MockSignalBackend` backend.
    :param callback: The callback to call when the backend is put to during the
        context.
    """
    backend = _get_mock_signal_backend(signal)
    backend.set_mock_put_callback(callback)
    return _unset_side_effect_cm(backend)


def set_mock_put_proceeds(signal: Signal, proceeds: bool):
    """Allow or block a put with wait=True from proceeding."""
    backend = _get_mock_signal_backend(signal)

    if proceeds:
        backend.put_proceeds.set()
    else:
        backend.put_proceeds.clear()


@contextmanager
def mock_puts_blocked(*signals: Signal):
    """Context manager to block puts at the start and unblock at the end."""
    for signal in signals:
        set_mock_put_proceeds(signal, False)
    yield
    for signal in signals:
        set_mock_put_proceeds(signal, True)


def get_mock_put(signal: Signal) -> AsyncMock:
    """Get the mock associated with the put call on the signal."""
    return _get_mock_signal_backend(signal).put_mock



from epicscorelibs.ca import dbr
from p4p import Value as P4PValue
from p4p.nt import NTEnum

from ophyd_async.core import SubsetEnum
from ophyd_async.epics.core import epics_signal_rw

# Allow these imports from private modules for tests
from ophyd_async.epics.core._aioca import (  # noqa: PLC2701
    make_converter as ca_make_converter,
)
from ophyd_async.epics.core._p4p import (  # noqa: PLC2701
    make_converter as pva_make_converter,
)


class AB(SubsetEnum):
    A = "A"
    B = "B"


class AB1(SubsetEnum):
    A = "A1"
    B = "B1"


class AB2(SubsetEnum):
    A = "A2"
    B = "B2"


async def test_ca_runtime_enum_converter():
    class EpicsValue:
        def __init__(self):
            self.name = "test"
            self.ok = (True,)
            self.errorcode = 0
            self.datatype = dbr.DBR_ENUM
            self.element_count = 1
            self.severity = 0
            self.status = 0
            self.raw_stamp = (0,)
            self.timestamp = 0
            self.datetime = 0
            self.enums = ["A", "B", "C"]  # More than the runtime enum

    epics_value = EpicsValue()
    converter = ca_make_converter(
        AB, values={"READ_PV": epics_value, "WRITE_PV": epics_value}
    )
    assert converter.supported_values == {"A": "A", "B": "B", "C": "C"}
    assert set(AB).issubset(set(converter.supported_values.keys()))


async def test_pva_runtime_enum_converter():
    enum_type = NTEnum.buildType()
    epics_value = P4PValue(
        enum_type,
        {
            "value.choices": ["A", "B", "C"],
        },
    )
    converter = pva_make_converter(
        AB, values={"READ_PV": epics_value, "WRITE_PV": epics_value}
    )
    assert {"A", "B"}.issubset(set(converter.supported_values))


async def test_runtime_enum_signal():
    signal_rw_pva = epics_signal_rw(AB1, "ca://RW_PV", name="signal")
    signal_rw_ca = epics_signal_rw(AB2, "ca://RW_PV", name="signal")
    await signal_rw_pva.connect(mock=True)
    await signal_rw_ca.connect(mock=True)
    assert await signal_rw_pva.get_value() == "A1"
    assert await signal_rw_ca.get_value() == "A2"
    await signal_rw_pva.set("B1")
    await signal_rw_ca.set("B2")
    assert await signal_rw_pva.get_value() == "B1"
    assert await signal_rw_ca.get_value() == "B2"

    # Will accept string values even if they're not in the runtime enum
    # Though type checking should compain
    await signal_rw_pva.set("C1")  # type: ignore
    await signal_rw_ca.set("C2")  # type: ignore



import logging
import re
from unittest.mock import call

import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    DetectorTrigger,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
    callback_on_mock_put,
    get_mock,
    init_devices,
    set_mock_value,
)
from ophyd_async.fastcs.jungfrau import (
    AcquisitionType,
    JungfrauDetector,
    JungfrauTriggerMode,
    PedestalMode,
)
from ophyd_async.testing import assert_has_calls


@pytest.fixture
def jungfrau(RE: RunEngine, tmp_path):
    path_provider = StaticPathProvider(StaticFilenameProvider("filename"), tmp_path)
    with init_devices(mock=True):
        detector = JungfrauDetector("prefix", path_provider, "", "", "jungfrau")

    # Enough to satisfy the odin writer
    set_mock_value(detector.odin.fp.writing, True)
    set_mock_value(detector.odin.mw.writing, True)
    set_mock_value(detector.detector.bit_depth, 8)
    return detector


@pytest.mark.parametrize(
    "trigger_info, match",
    [
        (
            TriggerInfo(trigger=DetectorTrigger.EXTERNAL_LEVEL),
            "Trigger type DetectorTrigger.EXTERNAL_LEVEL not supported by 'jungfrau', "
            "supported types are: [EXTERNAL_EDGE, INTERNAL]",
        ),
        (
            TriggerInfo(livetime=8e-9, deadtime=8e-9),
            "Period between frames (exposure time + deadtime) = 1.6e-08s "
            "cannot be lower than minimum detector deadtime 2e-05",
        ),
    ],
)
async def test_prepare_val_error_on_bad_trigger_info(
    trigger_info: TriggerInfo, match: str, jungfrau: JungfrauDetector
):
    with pytest.raises(
        ValueError,
        match=re.escape(match),
    ):
        await jungfrau.prepare(trigger_info)


async def test_prepare_warn_on_small_exposure(jungfrau: JungfrauDetector, caplog):
    bad_trigger_info = TriggerInfo(livetime=1e-6)
    with caplog.at_level(logging.WARNING):
        await jungfrau.prepare(bad_trigger_info)
    assert "Exposure time shorter than 2μs is not recommended" in caplog.messages


async def test_arm(jungfrau: JungfrauDetector):
    mock = get_mock(jungfrau.detector)
    await jungfrau.prepare(TriggerInfo())
    mock.reset_mock()
    callback_on_mock_put(
        jungfrau.detector.acquisition_start,
        lambda v: set_mock_value(jungfrau.odin.fp.frames_written, 1),
    )
    await jungfrau.trigger()
    assert_has_calls(jungfrau.detector, [call.acquisition_start.put(None)])


async def test_disarm(jungfrau: JungfrauDetector):
    await jungfrau.unstage()
    assert_has_calls(
        jungfrau.detector,
        [
            call.acquisition_stop.put(None),
            call.pedestal_mode_state.put(PedestalMode.OFF),
        ],
    )


async def test_signals_set_in_pedestal_mode(jungfrau: JungfrauDetector):
    await jungfrau.acquisition_type.set(AcquisitionType.PEDESTAL)
    await jungfrau.detector.pedestal_mode_frames.set(5)
    await jungfrau.detector.pedestal_mode_loops.set(3)
    good_trigger_info = TriggerInfo(
        livetime=1e-3,
        # We make 2 * frames * loops collections
        collections_per_event=2 * 5 * 3,
        trigger=DetectorTrigger.INTERNAL,
    )
    mock = get_mock(jungfrau.detector)
    mock.reset_mock()
    await jungfrau.prepare(good_trigger_info)
    assert_has_calls(
        jungfrau.detector,
        [
            call.trigger_mode.put(JungfrauTriggerMode.INTERNAL),
            call.period_between_frames.put(0.00102),
            call.exposure_time.put(0.001),
            call.pedestal_mode_state.put(PedestalMode.ON),
        ],
    )


async def test_signals_set_in_standard_internal_mode(jungfrau: JungfrauDetector):
    good_trigger_info = TriggerInfo(
        livetime=1e-3,
        collections_per_event=10,
        trigger=DetectorTrigger.INTERNAL,
    )
    await jungfrau.prepare(good_trigger_info)
    assert_has_calls(
        jungfrau.detector,
        [
            call.trigger_mode.put(JungfrauTriggerMode.INTERNAL),
            call.frames_per_acq.put(10),
            call.period_between_frames.put(0.00102),
            call.exposure_time.put(0.001),
        ],
    )


async def test_signals_set_in_standard_external_mode(jungfrau: JungfrauDetector):
    good_trigger_info = TriggerInfo(
        livetime=1e-3,
        number_of_events=10,
        trigger=DetectorTrigger.EXTERNAL_EDGE,
        deadtime=1,
    )
    await jungfrau.prepare(good_trigger_info)
    assert_has_calls(
        jungfrau.detector,
        [
            call.trigger_mode.put(JungfrauTriggerMode.EXTERNAL),
            call.frames_per_acq.put(10),
            call.period_between_frames.put(0.00102),
            call.exposure_time.put(0.001),
            call.acquisition_start.put(None),
        ],
    )


async def test_error_in_pedestal_and_external_modes(jungfrau: JungfrauDetector):
    await jungfrau.acquisition_type.set(AcquisitionType.PEDESTAL)
    with pytest.raises(
        ValueError,
        match="Jungfrau must be triggered internally while in pedestal mode.",
    ):
        await jungfrau.prepare(
            TriggerInfo(
                livetime=1e-3, trigger=DetectorTrigger.EXTERNAL_EDGE, deadtime=1
            )
        )


async def test_prepare_val_error_if_pedestal_mode_and_odd_number_of_events(
    jungfrau: JungfrauDetector,
):
    # No. events = pedestal loops * 2, so it should always be even
    await jungfrau.acquisition_type.set(AcquisitionType.PEDESTAL)
    await jungfrau.detector.pedestal_mode_frames.set(5)
    await jungfrau.detector.pedestal_mode_loops.set(3)

    trigger_info = TriggerInfo(livetime=1e-3, deadtime=1, collections_per_event=15)

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Invalid trigger info for pedestal mode. "
            "Number 15 must be equal to 2 * 5 * 3. "
            "Was create_jungfrau_pedestal_triggering_info used?",
        ),
    ):
        await jungfrau.prepare(trigger_info)


async def test_deadtime(jungfrau: JungfrauDetector):
    supported_triggers, deadtime = await jungfrau.get_trigger_deadtime()
    assert deadtime == 2e-5
    assert supported_triggers == {
        DetectorTrigger.EXTERNAL_EDGE,
        DetectorTrigger.INTERNAL,
    }



import asyncio
import time

import numpy as np
from tango import AttrWriteType, DevState, GreenMode
from tango.server import Device, attribute, command


class DemoMover(Device):
    """Demo tango moving device."""

    green_mode = GreenMode.Asyncio
    _position = 0.0
    _setpoint = 0.0
    _velocity = 0.5
    _acceleration = 0.5
    _precision = 0.1
    _stop = False
    DEVICE_CLASS_INITIAL_STATE = DevState.ON

    @attribute(dtype=float, access=AttrWriteType.READ_WRITE)
    async def position(self):
        return self._position

    async def write_position(self, new_position):
        self._setpoint = new_position
        await self.move()

    @attribute(dtype=float, access=AttrWriteType.READ_WRITE)
    async def velocity(self):
        return self._velocity

    async def write_velocity(self, value: float):
        self._velocity = value

    @attribute(dtype=DevState, access=AttrWriteType.READ)
    async def state(self):
        return self.get_state()

    @command
    async def stop(self):
        self._stop = True

    @command
    async def move(self):
        self.set_state(DevState.MOVING)
        await self._move(self._setpoint)
        self.set_state(DevState.ON)

    async def _move(self, new_position):
        self._setpoint = new_position
        self._stop = False
        step = 0.1
        while True:
            if self._stop:
                self._stop = False
                break
            if self._position < new_position:
                self._position = self._position + self._velocity * step
            else:
                self._position = self._position - self._velocity * step
            if abs(self._position - new_position) < self._precision:
                self._position = new_position
                break
            await asyncio.sleep(step)


class DemoCounter(Device):
    """Demo tango counting device."""

    green_mode = GreenMode.Asyncio
    _counts = 0
    _sample_time = 1.0

    @attribute(dtype=int, access=AttrWriteType.READ)
    async def counts(self):
        return self._counts

    @attribute(dtype=float, access=AttrWriteType.READ_WRITE)
    async def sample_time(self):
        return self._sample_time

    async def write_sample_time(self, value: float):
        self._sample_time = value

    @attribute(dtype=DevState, access=AttrWriteType.READ)
    async def state(self):
        return self.get_state()

    @command
    async def reset(self):
        self._counts = 0
        return self._counts

    @command
    async def start(self):
        self._counts = 0
        if self._sample_time <= 0.0:
            return
        self.set_state(DevState.MOVING)
        await self._trigger()
        self.set_state(DevState.ON)

    async def _trigger(self):
        st = time.time()
        while True:
            ct = time.time()
            if ct - st > self._sample_time:
                break
            self._counts += int(np.random.normal(1000, 100))
            await asyncio.sleep(0.1)



from ophyd_async.core import PathProvider, SignalR, StandardDetector
from ophyd_async.fastcs import odin
from ophyd_async.fastcs.core import fastcs_connector

from ._arm_logic import EigerArmLogic
from ._io import EigerDetectorIO, EigerMonitorIO, EigerStreamIO
from ._trigger_logic import EigerTriggerLogic


class EigerDetector(StandardDetector):
    """Ophyd-async implementation of an Eiger Detector."""

    stale_parameters: SignalR[bool]
    monitor: EigerMonitorIO
    stream: EigerStreamIO
    detector: EigerDetectorIO
    od: odin.OdinIO

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        name="",
    ):
        # Need to do this first so the type hints are filled in
        connector = fastcs_connector(prefix, self)
        self.add_detector_logics(
            EigerTriggerLogic(self.detector),
            EigerArmLogic(self.detector),
            odin.OdinDataLogic(
                path_provider=path_provider,
                odin=self.od,
                detector_bit_depth=self.detector.bit_depth_image,
            ),
        )
        super().__init__(name=name, connector=connector)



"""FastCS core module for ophyd-async."""

from ophyd_async.core import Device, DeviceConnector
from ophyd_async.epics.core import PviDeviceConnector


def fastcs_connector(
    uri: str, device: Device | None = None, error_hint: str = ""
) -> DeviceConnector:
    """Create devices and connections on pvi device `Device`."""
    # TODO: add Tango support based on uri scheme
    connector = PviDeviceConnector(uri, error_hint)
    if device:
        connector.create_children_from_annotations(device)
    return connector


__all__ = ["fastcs_connector"]



import asyncio
import time

import numpy as np

from ophyd_async.core import (
    AsyncStatus,
    DeviceVector,
    SignalR,
    StandardReadable,
    StrictEnum,
    gather_dict,
    soft_signal_r_and_setter,
    soft_signal_rw,
)
from ophyd_async.core import StandardReadableFormat as Format

from ._pattern_generator import PatternGenerator


class EnergyMode(StrictEnum):
    """Energy mode for `SimPointDetector`."""

    LOW = "Low Energy"
    """Low energy mode"""

    HIGH = "High Energy"
    """High energy mode"""


class SimPointDetectorChannel(StandardReadable):
    def __init__(self, value_signal: SignalR[int], name=""):
        with self.add_children_as_readables(Format.HINTED_SIGNAL):
            self.value = value_signal
        with self.add_children_as_readables(Format.CONFIG_SIGNAL):
            self.mode = soft_signal_rw(EnergyMode)
        super().__init__(name)


class SimPointDetector(StandardReadable):
    """Simalutes a point detector with multiple channels."""

    def __init__(
        self,
        pattern_generator: PatternGenerator | None,
        num_channels: int = 3,
        name: str = "",
    ) -> None:
        self.pattern_generator = pattern_generator or PatternGenerator()
        self.acquire_time = soft_signal_rw(float, 0.1)
        self.acquiring, self._set_acquiring = soft_signal_r_and_setter(bool)
        self._value_signals = dict(
            soft_signal_r_and_setter(int) for _ in range(num_channels)
        )
        with self.add_children_as_readables():
            self.channel = DeviceVector(
                {
                    i + 1: SimPointDetectorChannel(value_signal)
                    for i, value_signal in enumerate(self._value_signals)
                }
            )
        super().__init__(name=name)

    async def _update_values(self, acquire_time: float):
        # Get the modes
        modes = await gather_dict(
            {channel: channel.mode.get_value() for channel in self.channel.values()}
        )
        start = time.monotonic()
        # Make an array of relative update times at 10Hz intervals
        update_times = np.arange(0.1, acquire_time, 0.1)
        # With the end position appended
        update_times = np.concatenate((update_times, [acquire_time]))
        for update_time in update_times:
            # Calculate how long to wait to get there
            relative_time = time.monotonic() - start
            await asyncio.sleep(update_time - relative_time)
            # Update the channel value
            for i, channel in self.channel.items():
                high_energy = modes[channel] == EnergyMode.HIGH
                point = self.pattern_generator.generate_point(i, high_energy)
                setter = self._value_signals[channel.value]
                setter(int(point * 10000 * update_time))

    @AsyncStatus.wrap
    async def trigger(self):
        for setter in self._value_signals.values():
            setter(0)
        await self._update_values(await self.acquire_time.get_value())



from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DetectorArmLogic,
    SignalR,
    set_and_wait_for_other_value,
    set_and_wait_for_value,
)
from ophyd_async.epics.core import stop_busy_record, wait_for_good_state

from ._io import ADBaseIO, ADState, NDCircularBuffIO


class ADArmLogic(DetectorArmLogic):
    def __init__(
        self, driver: ADBaseIO, driver_armed_signal: SignalR[bool] | None = None
    ):
        self.driver = driver
        if driver_armed_signal is not None:
            self.driver_armed_signal = driver_armed_signal
        else:
            self.driver_armed_signal = driver.acquire
        self.acquire_status: AsyncStatus | None = None

    async def arm(self):
        self.acquire_status = await set_and_wait_for_other_value(
            set_signal=self.driver.acquire,
            set_value=True,
            match_signal=self.driver_armed_signal,
            match_value=True,
            wait_for_set_completion=False,
            timeout=DEFAULT_TIMEOUT,
        )

    async def wait_for_idle(self):
        if self.acquire_status:
            await self.acquire_status
        await wait_for_good_state(
            self.driver.detector_state,
            {ADState.IDLE, ADState.ABORTED},
            timeout=DEFAULT_TIMEOUT,
        )

    async def disarm(self):
        await stop_busy_record(self.driver.acquire)


class ADContAcqArmLogic(DetectorArmLogic):
    def __init__(self, driver: ADBaseIO, cb_plugin: NDCircularBuffIO):
        self.driver = driver
        self.cb_plugin = cb_plugin
        self.acquire_status: AsyncStatus | None = None

    async def arm(self):
        self.acquire_status = await set_and_wait_for_value(
            self.cb_plugin.capture,
            True,
            wait_for_set_completion=False,
            timeout=DEFAULT_TIMEOUT,
        )

    async def wait_for_idle(self):
        if self.acquire_status:
            await self.acquire_status

    async def disarm(self):
        await stop_busy_record(self.cb_plugin.capture)



import pytest

from ophyd_async.epics import adcore


def test_get_plugin_type_checking():
    driver = adcore.ADBaseIO("PREFIX:DRV:")
    plugins = {
        "stats": adcore.NDStatsIO("PREFIX:STAT:"),
    }
    det = adcore.AreaDetector(
        driver=driver, writer_type=None, plugins=plugins, name="det"
    )

    # Correct type
    stats_plugin = det.get_plugin("stats", adcore.NDStatsIO)
    assert isinstance(stats_plugin, adcore.NDStatsIO)

    # Incorrect type
    with pytest.raises(
        TypeError, match=r"Expected det\.stats to be a .*NDPluginFileIO.*"
    ):
        det.get_plugin("stats", adcore.NDPluginFileIO)



import importlib
import re
import time
from pathlib import Path
from unittest.mock import patch

import pytest
from aioca import purge_channel_caches
from bluesky.run_engine import RunEngine

# https://regex101.com/r/KvLj7t/1
SCAN_LINE = re.compile(
    r"^\| *(\d+) \|[^\|]*\| *(\d*.\d*) \| *(\d*.\d*) \| *(\d*) \| *(\d*) \| *(\d*) \|$",
    re.M,
)


@pytest.fixture
def expected_scan_output():
    tutorial_text = (
        Path(__file__).absolute().parent.parent.parent
        / "docs"
        / "tutorials"
        / "implementing-devices.md"
    ).read_text()
    matches = SCAN_LINE.findall(tutorial_text)
    assert len(matches) == 9
    yield matches


@pytest.mark.parametrize("module", ["ophyd_async.sim", "ophyd_async.epics.demo"])
def test_implementing_devices(module, capsys, expected_scan_output):
    with patch("matplotlib.get_backend"):
        with patch("bluesky.run_engine.autoawait_in_bluesky_event_loop") as autoawait:
            main = importlib.import_module(f"{module}.__main__")
            autoawait.assert_called_once_with()
        # We want the text output of the best effort callback, but the plotting takes
        # too much time for CI, even if headless, so disable it
        main.bec._set_up_plots = lambda *args, **kwargs: None
        RE: RunEngine = main.RE
        for motor in [main.stage.x, main.stage.y]:
            RE(main.bps.mv(motor.velocity, 1000))
        start = time.monotonic()
        RE(main.bp.grid_scan([main.pdet], main.stage.x, 1, 2, 3, main.stage.y, 2, 3, 3))
        assert time.monotonic() - start == pytest.approx(2.5, abs=2.0)
        captured = capsys.readouterr()
        assert captured.err == ""
        assert SCAN_LINE.findall(captured.out) == expected_scan_output

        # If we are testing the EPICS demo, we need to stop the IOC and purge caches
        # to avoid CA virtual circuit disconnect errors.
        if module == "ophyd_async.epics.demo":
            purge_channel_caches()
            main.ioc.stop()



import asyncio
import math
from typing import TypedDict

from bluesky.protocols import Movable

from ophyd_async.core import (
    AsyncStatus,
    DerivedSignalFactory,
    Device,
    Transform,
    soft_signal_rw,
)

from ._motor import SimMotor


class TwoJackRaw(TypedDict):
    jack1: float
    jack2: float


class TwoJackDerived(TypedDict):
    height: float
    angle: float


class TwoJackTransform(Transform):
    distance: float

    def raw_to_derived(self, *, jack1: float, jack2: float) -> TwoJackDerived:
        diff = jack2 - jack1
        return TwoJackDerived(
            height=jack1 + diff / 2,
            # need the cast as returns numpy float rather than float64, but this
            # is ok at runtime
            angle=math.atan(diff / self.distance),
        )

    def derived_to_raw(self, *, height: float, angle: float) -> TwoJackRaw:
        diff = math.tan(angle) * self.distance
        return TwoJackRaw(
            jack1=height - diff / 2,
            jack2=height + diff / 2,
        )


class VerticalMirror(Device, Movable[TwoJackDerived]):
    def __init__(self, name=""):
        # Raw signals
        self.y1 = SimMotor()
        self.y2 = SimMotor()
        # Parameter
        # This could also be set as '1.0', if constant.
        self.y1_y2_distance = soft_signal_rw(float, initial_value=1)
        # Derived signals
        self._factory = DerivedSignalFactory(
            TwoJackTransform,
            self.set,
            jack1=self.y1,
            jack2=self.y2,
            distance=self.y1_y2_distance,
        )
        self.height = self._factory.derived_signal_rw(float, "height")
        self.angle = self._factory.derived_signal_rw(float, "angle")
        super().__init__(name=name)

    @AsyncStatus.wrap
    async def set(self, derived: TwoJackDerived) -> None:  # type: ignore until bluesky 1.13.2
        transform = await self._factory.transform()
        raw = transform.derived_to_raw(**derived)
        await asyncio.gather(
            self.y1.set(raw["jack1"]),
            self.y2.set(raw["jack2"]),
        )



import asyncio
import contextlib
import time

import numpy as np
from bluesky.protocols import Locatable, Location, Reading, Stoppable, Subscribable

from ophyd_async.core import (
    AsyncStatus,
    Callback,
    FlyMotorInfo,
    StandardReadable,
    WatchableAsyncStatus,
    WatcherUpdate,
    error_if_none,
    observe_value,
    soft_signal_r_and_setter,
    soft_signal_rw,
)
from ophyd_async.core import StandardReadableFormat as Format


class SimMotor(StandardReadable, Stoppable, Subscribable[float], Locatable[float]):
    """For usage when simulating a motor."""

    def __init__(
        self,
        name: str = "",
        instant: bool = True,
        initial_value: float = 0.0,
        units: str = "mm",
    ) -> None:
        """Simulate a motor, with optional velocity.

        :param name: name of device
        :param instant: whether to move instantly or calculate move time using velocity
        :param initial_value: initial position of the motor
        :param units: units of the motor position
        """
        # Define some signals
        with self.add_children_as_readables(Format.HINTED_SIGNAL):
            self.user_readback, self._user_readback_set = soft_signal_r_and_setter(
                float, 0
            )
        with self.add_children_as_readables(Format.CONFIG_SIGNAL):
            self.velocity = soft_signal_rw(float, 0 if instant else 1.0)
            self.acceleration_time = soft_signal_rw(float, 0.5)
            self.units = soft_signal_rw(str, units)
        self.user_setpoint = soft_signal_rw(float, initial_value)

        # Whether set() should complete successfully or not
        self._set_success = True
        self._move_status: AsyncStatus | None = None
        # Stored in prepare
        self._fly_info: FlyMotorInfo | None = None
        # Set on kickoff(), complete when motor reaches end position
        self._fly_status: WatchableAsyncStatus | None = None

        super().__init__(name=name)

    def set_name(self, name: str, *, child_name_separator: str | None = None) -> None:
        super().set_name(name, child_name_separator=child_name_separator)
        # Readback should be named the same as its parent in read()
        self.user_readback.set_name(name)

    @AsyncStatus.wrap
    async def prepare(self, value: FlyMotorInfo):
        """Calculate run-up and move there, setting fly velocity when there."""
        self._fly_info = value
        # Move to start as fast as we can
        await self.velocity.set(0)
        await self.set(
            value.ramp_up_start_pos(await self.acceleration_time.get_value())
        )
        # Set the velocity for the actual move
        await self.velocity.set(value.velocity)

    async def locate(self) -> Location[float]:
        """Return the current setpoint and readback of the motor."""
        setpoint, readback = await asyncio.gather(
            self.user_setpoint.get_value(), self.user_readback.get_value()
        )
        return Location(setpoint=setpoint, readback=readback)

    def subscribe_reading(self, function: Callback[dict[str, Reading[float]]]) -> None:
        self.user_readback.subscribe_reading(function)

    subscribe = subscribe_reading

    def clear_sub(self, function: Callback[dict[str, Reading[float]]]) -> None:
        self.user_readback.clear_sub(function)

    @AsyncStatus.wrap
    async def kickoff(self):
        """Begin moving motor from prepared position to final position."""
        fly_info = error_if_none(
            self._fly_info, "Motor must be prepared before attempting to kickoff"
        )
        acceleration_time = await self.acceleration_time.get_value()
        self._fly_status = self.set(fly_info.ramp_down_end_pos(acceleration_time))
        # Wait for the acceleration time to ensure we are at velocity
        await asyncio.sleep(acceleration_time)

    def complete(self) -> WatchableAsyncStatus:
        """Mark as complete once motor reaches completed position."""
        fly_status = error_if_none(self._fly_status, "kickoff not called")
        return fly_status

    async def _move(self, old_position: float, new_position: float, velocity: float):
        if old_position == new_position:
            return
        start = time.monotonic()
        acceleration_time = abs(await self.acceleration_time.get_value())
        sign = np.sign(new_position - old_position)
        velocity = abs(velocity) * sign
        # The total distance to move
        total_distance = new_position - old_position
        # The ramp distance is the distance taken to ramp up (the same distance
        # is taken to ramp down). This is the area under the triangle of the
        # velocity ramp up (base * height / 2)
        ramp_distance = acceleration_time * velocity / 2
        if abs(ramp_distance * 2) >= abs(total_distance):
            # All time is ramp up and down, so recalculate the maximum velocity
            # we get to. We know the area under the ramp up triangle is half the
            # total distance, and we also know the ratio of velocity over
            # acceleration_time is the same as the ration of max_velocity over
            # ramp_time, so solve the simultaneous equations to get
            # max_velocity and ramp_time.
            max_velocity = np.sqrt(total_distance * velocity / acceleration_time) * sign
            ramp_time = total_distance / max_velocity
            # So move time is just the ramp up and ramp down with no constant
            # velocity section
            move_time = 2 * ramp_time
        else:
            # Middle segments of constant velocity
            max_velocity = velocity
            # Ramp up and down time is exactly the requested acceleration time
            ramp_time = acceleration_time
            # So move time is twice this, plus the time taken to move the
            # remaining distance at constant velocity
            move_time = ramp_time * 2 + (total_distance - ramp_distance * 2) / velocity
        # Make an array of relative update times at 10Hz intervals
        update_times = list(np.arange(0.1, move_time, 0.1, dtype=float))
        # With the end position appended
        if update_times and np.isclose(update_times[-1], move_time):
            update_times[-1] = move_time
        else:
            update_times.append(move_time)
        # Iterate through the update times, calculating new position for each
        for t in update_times:
            if t <= ramp_time:
                # Ramp up phase, calculate area under the ramp up triangle
                current_velocity = t / ramp_time * max_velocity
                position = old_position + current_velocity * t / 2
            elif t >= move_time - ramp_time:
                # Ramp down phase, subtract area under the ramp down triangle
                time_left = move_time - t
                current_velocity = time_left / ramp_time * max_velocity
                position = new_position - current_velocity * time_left / 2
            else:
                # Constant velocity phase
                position = old_position + ramp_distance + (t - ramp_time) * max_velocity
            # Calculate how long to wait to get there
            relative_time = time.monotonic() - start
            await asyncio.sleep(t - relative_time)
            # Update the readback position
            self._user_readback_set(position)

    @WatchableAsyncStatus.wrap
    async def set(self, value: float):
        """Asynchronously move the motor to a new position."""
        new_position = value
        # Make sure any existing move tasks are stopped
        if self._move_status:
            self._move_status.task.cancel()
            self._move_status = None
        # work out where we were
        old_position, units, velocity = await asyncio.gather(
            self.user_setpoint.get_value(),
            self.units.get_value(),
            self.velocity.get_value(),
        )
        # update the setpoint to where we want to be
        await self.user_setpoint.set(new_position)
        # If zero velocity, do instant move
        if velocity == 0:
            self._user_readback_set(new_position)
        else:
            self._move_status = AsyncStatus(
                self._move(old_position, new_position, velocity)
            )
            # If stop is called then this will raise a CancelledError, ignore it
            with contextlib.suppress(asyncio.CancelledError):
                async for current_position in observe_value(
                    self.user_readback, done_status=self._move_status
                ):
                    yield WatcherUpdate(
                        current=current_position,
                        initial=old_position,
                        target=new_position,
                        name=self.name,
                        unit=units,
                    )
        if not self._set_success:
            raise RuntimeError("Motor was stopped")

    async def stop(self, success=True):
        """Stop the motor if it is moving."""
        self._set_success = success
        if self._move_status:
            self._move_status.task.cancel()
            self._move_status = None
        await self.user_setpoint.set(await self.user_readback.get_value())



"""Helpful config class for testing."""

from tango import AttrDataFormat, CmdArgType


class TestConfig:
    """Configuration for a test."""

    data_type: CmdArgType
    data_format: AttrDataFormat
    enum_labels: list[str]



import numpy as np
import yaml
from bluesky import RunEngine

from ophyd_async.core import YamlSettingsProvider, init_devices
from ophyd_async.epics.core import epics_signal_rw
from ophyd_async.fastcs.core import fastcs_connector
from ophyd_async.fastcs.panda import (
    CommonPandaBlocks,
    DataBlock,
    PandaTimeUnits,
    SeqTable,
    apply_panda_settings,
)
from ophyd_async.plan_stubs import retrieve_settings, store_settings


async def get_mock_panda():
    class Panda(CommonPandaBlocks):
        data: DataBlock

        def __init__(self, uri: str, name: str = ""):
            super().__init__(name=name, connector=fastcs_connector(uri, self))

    async with init_devices(mock=True):
        mock_panda = Panda("PANDA")
        mock_panda.phase_1_signal_units = epics_signal_rw(int, "")
    return mock_panda


async def test_save_load_panda(tmp_path, RE: RunEngine):
    mock_panda1 = await get_mock_panda()
    await mock_panda1.seq[1].table.set(SeqTable.row(repeats=1))
    provider = YamlSettingsProvider(tmp_path)
    RE(store_settings(provider, "panda", mock_panda1))

    def check_equal_with_seq_tables(actual, expected):
        assert actual.model_fields_set == expected.model_fields_set
        for field_name, field_value1 in actual:
            field_value2 = getattr(expected, field_name)
            assert np.array_equal(field_value1, field_value2)

    mock_panda2 = await get_mock_panda()
    check_equal_with_seq_tables(
        (await mock_panda2.seq[1].table.get_value()),
        SeqTable(),  # type: ignore
    )

    def load_panda():
        settings = yield from retrieve_settings(provider, "panda", mock_panda2)
        yield from apply_panda_settings(settings)

    RE(load_panda())

    check_equal_with_seq_tables(
        await mock_panda2.seq[1].table.get_value(),
        SeqTable.row(repeats=1),
    )

    # Load the YAML content as a string
    with open(str(tmp_path / "panda.yaml")) as file:
        yaml_content = file.read()

    # Parse the YAML content
    parsed_yaml = yaml.safe_load(yaml_content)

    assert parsed_yaml == {
        "phase_1_signal_units": 0,
        "seq.1.prescale_units": PandaTimeUnits("min"),
        "seq.2.prescale_units": PandaTimeUnits("min"),
        "data.capture": False,
        "data.capture_mode": "FIRST_N",
        "data.create_directory": 0,
        "data.flush_period": 0.0,
        "data.hdf_directory": "",
        "data.hdf_file_name": "",
        "data.num_capture": 0,
        "pcap.arm": False,
        "pcomp.1.dir": "Positive",
        "pcomp.1.enable": "ZERO",
        "pcomp.1.pulses": 0,
        "pcomp.1.start": 0,
        "pcomp.1.step": 0,
        "pcomp.1.width": 0,
        "pcomp.2.dir": "Positive",
        "pcomp.2.enable": "ZERO",
        "pcomp.2.pulses": 0,
        "pcomp.2.start": 0,
        "pcomp.2.step": 0,
        "pcomp.2.width": 0,
        "pulse.1.delay": 0.0,
        "pulse.1.width": 0.0,
        "pulse.1.step": 0.0,
        "pulse.1.pulses": 0,
        "pulse.1.enable": "ZERO",
        "pulse.2.delay": 0.0,
        "pulse.2.width": 0.0,
        "pulse.2.step": 0.0,
        "pulse.2.pulses": 0,
        "pulse.2.enable": "ZERO",
        "seq.1.table": {
            "outa1": [False],
            "outa2": [False],
            "outb1": [False],
            "outb2": [False],
            "outc1": [False],
            "outc2": [False],
            "outd1": [False],
            "outd2": [False],
            "oute1": [False],
            "oute2": [False],
            "outf1": [False],
            "outf2": [False],
            "position": [0],
            "repeats": [1],
            "time1": [0],
            "time2": [0],
            "trigger": ["Immediate"],
        },
        "seq.1.posa": "ZERO",
        "seq.1.repeats": 0,
        "seq.1.prescale": 0.0,
        "seq.1.enable": "ZERO",
        "seq.2.table": {
            "outa1": [],
            "outa2": [],
            "outb1": [],
            "outb2": [],
            "outc1": [],
            "outc2": [],
            "outd1": [],
            "outd2": [],
            "oute1": [],
            "oute2": [],
            "outf1": [],
            "outf2": [],
            "position": [],
            "repeats": [],
            "time1": [],
            "time2": [],
            "trigger": [],
        },
        "seq.2.posa": "ZERO",
        "seq.2.repeats": 0,
        "seq.2.prescale": 0.0,
        "seq.2.enable": "ZERO",
    }



import asyncio


async def wait_for_pending_wakeups(max_yields=20, raise_if_exceeded=True):
    """Allow any ready asyncio tasks to be woken up.

    Used in:

    - Tests to allow tasks like ``set()`` to start so that signal
      puts can be tested
    - `observe_value` to allow it to be wrapped in `asyncio.wait_for`
      with a timeout
    """
    loop = asyncio.get_event_loop()
    # If anything has called loop.call_soon or is scheduled a wakeup
    # then let it run
    for _ in range(max_yields):
        await asyncio.sleep(0)
        if not loop._ready:  # type: ignore # noqa: SLF001
            return
    if raise_if_exceeded:
        raise RuntimeError(f"Tasks still scheduling wakeups after {max_yields} yields")



from __future__ import annotations

import asyncio
import re
from collections.abc import Sequence
from dataclasses import dataclass

import numpy as np
from scanspec.core import Slice

from ophyd_async.core import gather_dict
from ophyd_async.epics.motor import Motor

from ._pmac_io import CS_INDEX, PmacIO

# PMAC durations are in milliseconds
# We must convert from scanspec durations (seconds) to milliseconds
# PMAC motion program multiples durations by 0.001
# (see https://github.com/DiamondLightSource/pmac/blob/afe81f8bb9179c3a20eff351f30bc6cfd1539ad9/pmacApp/pmc/trajectory_scan_code_ppmac.pmc#L241)
# Therefore, we must divide scanspec durations by 10e-6
TICK_S = 0.000001
MIN_TURNAROUND = 0.002
MIN_INTERVAL = 0.002

# Regex to parse outlink strings of the form "@asyn(PMAC1CS2, 7)"
# returning PMAC1CS2 and 7
# https://regex101.com/r/Mu9XpO/1
OUTLINK_REGEX = re.compile(r"^\@asyn\(([^,]+),\s*(\d+)\)$")


@dataclass
class _PmacMotorInfo:
    cs_port: str
    cs_number: int
    motor_cs_index: dict[Motor, int]
    motor_acceleration_rate: dict[Motor, float]
    motor_max_velocity: dict[Motor, float]

    @classmethod
    async def from_motors(cls, pmac: PmacIO, motors: Sequence[Motor]) -> _PmacMotorInfo:
        """Creates a _PmacMotorInfo instance based on a  controller and list of motors.

        :param pmac: The PMAC device
        :param motors: Sequence of motors involved in trajectory
        :raises RuntimeError:
            if motors do not share common CS port or CS number, or if
            motors do not have unique CS index assignments
        :returns:
            _PmacMotorInfo instance with motor's common CS port and CS number, and
            dictionaries of motor's to their unique CS index and accelerate rate

        """
        is_raw_motor = [motor in pmac.motor_assignment_index for motor in motors]
        if all(is_raw_motor):
            # Get the CS port, number and axis letter from the PVs for the raw motor
            assignments = {
                motor: pmac.assignment[pmac.motor_assignment_index[motor]]
                for motor in motors
            }
            cs_ports, cs_numbers, cs_axis_letters = await asyncio.gather(
                gather_dict(
                    {motor: assignments[motor].cs_port.get_value() for motor in motors}
                ),
                gather_dict(
                    {
                        motor: assignments[motor].cs_number.get_value()
                        for motor in motors
                    }
                ),
                gather_dict(
                    {
                        motor: assignments[motor].cs_axis_letter.get_value()
                        for motor in motors
                    }
                ),
            )
            # Translate axis letters to cs_index and check for duplicates
            motor_cs_index: dict[Motor, int] = {}
            for motor, cs_axis_letter in cs_axis_letters.items():
                if not cs_axis_letter:
                    raise ValueError(
                        f"Motor {motor.name} does not have an axis assignment."
                    )
                try:
                    # 1 indexed to match coord
                    index = CS_INDEX[cs_axis_letter]
                except KeyError as err:
                    raise ValueError(
                        f"Motor {motor.name} assigned to '{cs_axis_letter}' "
                        f"but must be assigned to one of '{','.join(CS_INDEX)}'"
                    ) from err
                if index in motor_cs_index.values():
                    raise ValueError(
                        f"Motor {motor.name} assigned to '{cs_axis_letter}' "
                        "but another motor is already assigned to this axis."
                    )
                motor_cs_index[motor] = index
        elif not any(is_raw_motor):
            # Get CS numbers from all the cs ports and output links for the CS motors
            output_links, cs_lookup = await asyncio.gather(
                gather_dict({motor: motor.output_link.get_value() for motor in motors}),
                gather_dict(
                    {
                        cs_number: cs.cs_port.get_value()
                        for cs_number, cs in pmac.coord.items()
                    }
                ),
            )
            # Create a reverse lookup from cs_port to cs_number
            cs_reverse_lookup = {
                cs_port: cs_number for cs_number, cs_port in cs_lookup.items()
            }
            cs_ports: dict[Motor, str] = {}
            cs_numbers: dict[Motor, int] = {}
            motor_cs_index: dict[Motor, int] = {}
            # Populate the cs_ports, cs_numbers and motor_cs_index dicts from outlinks
            for motor, output_link in output_links.items():
                match = OUTLINK_REGEX.match(output_link)
                if not match:
                    raise ValueError(
                        f"Motor {motor.name} has invalid output link '{output_link}'"
                    )
                cs_port, cs_index = match.groups()
                cs_ports[motor] = cs_port
                cs_numbers[motor] = cs_reverse_lookup[cs_port]
                motor_cs_index[motor] = int(cs_index)
        else:
            raise ValueError("Unable to use raw motors and CS motors in the same scan")

        # check if the values in cs_port and cs_number are the same
        cs_ports_set = set(cs_ports.values())
        if len(cs_ports_set) != 1:
            raise RuntimeError(
                "Failed to fetch common CS port."
                "Motors passed are assigned to multiple CS ports:"
                f"{list(cs_ports_set)}"
            )
        cs_numbers_set = set(cs_numbers.values())
        if len(cs_numbers_set) != 1:
            raise RuntimeError(
                "Failed to fetch common CS number."
                "Motors passed are assigned to multiple CS numbers:"
                f"{list(cs_numbers_set)}"
            )

        # Get the velocities and acceleration rates for each motor
        max_velocity, acceleration_time = await asyncio.gather(
            gather_dict({motor: motor.max_velocity.get_value() for motor in motors}),
            gather_dict(
                {motor: motor.acceleration_time.get_value() for motor in motors}
            ),
        )
        motor_acceleration_rate = {
            motor: max_velocity[motor] / acceleration_time[motor] for motor in motors
        }
        return _PmacMotorInfo(
            cs_port=cs_ports_set.pop(),
            cs_number=cs_numbers_set.pop(),
            motor_cs_index=motor_cs_index,
            motor_acceleration_rate=motor_acceleration_rate,
            motor_max_velocity=max_velocity,
        )


def calculate_ramp_position_and_duration(
    slice: Slice[Motor], motor_info: _PmacMotorInfo, is_up: bool
) -> tuple[dict[Motor, np.float64], float]:
    """Calculate the the required ramp position and duration of a trajectory.

    This function will:
      - Calculate the ramp time required to achieve each motor's
        initial entry velocity into the first frame of a slice
        or final exit velocity out of the last frame of a slice
      - Use the longest ramp time to calculate all motor's
        ramp up positions.

    :param slice: Information about a series of scan frames along a number of axes
    :param motor_info: Instance of _PmacMotorInfo
    :param is_up: Boolean representing ramping up into a frame or down out of a frame
    :returns tuple: A tuple containing:
        dict: Motor to ramp positions
        float: Ramp time required for all motors
    """
    if slice.duration is None:
        raise ValueError("Slice must have a duration")

    scan_axes = slice.axes()
    idx = 0 if is_up else -1

    velocities: dict[Motor, float] = {}
    ramp_times: list[float] = []
    for axis in scan_axes:
        velocity = (slice.upper[axis][idx] - slice.lower[axis][idx]) / slice.duration[
            idx
        ]
        velocities[axis] = velocity
        ramp_times.append(abs(velocity) / motor_info.motor_acceleration_rate[axis])
    ramp_times.append(
        MIN_TURNAROUND
    )  # Adding a 2ms ramp time as a min tournaround time
    max_ramp_time = max(ramp_times)

    motor_to_ramp_position = {}
    sign = -1 if is_up else 1
    for axis, v in velocities.items():
        ref_pos = slice.lower[axis][0] if is_up else slice.upper[axis][-1]
        displacement = 0.5 * v * max_ramp_time
        motor_to_ramp_position[axis] = ref_pos + sign * displacement

    return (motor_to_ramp_position, max_ramp_time)



from unittest.mock import call

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    StaticFilenameProvider,
    StaticPathProvider,
    TriggerInfo,
    init_devices,
    set_mock_value,
)
from ophyd_async.fastcs.eiger import EigerDetector, EigerTriggerMode
from ophyd_async.testing import assert_has_calls


@pytest.fixture
def detector(RE, tmp_path):
    path_provider = StaticPathProvider(StaticFilenameProvider("filename"), tmp_path)
    with init_devices(mock=True):
        detector = EigerDetector("BL03I", path_provider)

    # Enough to satisfy the odin writer
    set_mock_value(detector.od.fp.writing, True)
    set_mock_value(detector.od.mw.writing, True)
    set_mock_value(detector.detector.bit_depth_image, 16)
    return detector


async def test_prepare_internal_calls_correct_parameters(
    detector: EigerDetector, tmp_path
):
    await detector.prepare(
        TriggerInfo(
            number_of_events=10,
            livetime=0.1,
            trigger=DetectorTrigger.INTERNAL,
        )
    )
    assert_has_calls(
        detector,
        [
            call.detector.trigger_mode.put(EigerTriggerMode.INTERNAL),
            call.detector.nimages.put(10),
            call.detector.count_time.put(0.1),
            call.detector.frame_time.put(0.1),
            call.od.fp.data_datatype.put("uint16"),
            call.od.fp.data_compression.put("BSLZ4"),
            call.od.fp.frames.put(0),
            call.od.fp.process_frames_per_block.put(1000),
            call.od.fp.file_path.put(str(tmp_path)),
            call.od.mw.directory.put(str(tmp_path)),
            call.od.fp.file_prefix.put("filename.h5"),
            call.od.mw.file_prefix.put("filename.h5"),
            call.od.mw.acquisition_id.put("filename.h5"),
            call.od.fp.start_writing.put(None),
        ],
        reset_after=False,
    )


async def test_deadtime_correct(detector: EigerDetector):
    supported_triggers, deadtime = await detector.get_trigger_deadtime()
    assert supported_triggers == {
        DetectorTrigger.INTERNAL,
        DetectorTrigger.EXTERNAL_EDGE,
        DetectorTrigger.EXTERNAL_LEVEL,
    }
    assert deadtime == 0.0001



"""Common GenICam camera deadtime values and utilities."""

__all__ = ["get_camera_deadtime"]

# Map model number to maximum deadtime in any pixel mode
# TODO: put back in the pixel format calcs from
# https://github.com/bluesky/ophyd-async/pull/190
_camera_deadtimes = {
    # cite: https://cdn.alliedvision.com/fileadmin/content/documents/products/cameras/Manta/techman/Manta_TechMan.pdf retrieved 2024-04-05  # noqa: E501
    "Manta G-125": 63e-6,
    "Manta G-145": 106e-6,
    "Manta G-235": 390e-6,
    "Manta G-895": 822e-6,
    "Manta G-2460": 1961e-6,
    # cite: https://cdn.alliedvision.com/fileadmin/content/documents/products/cameras/various/appnote/GigE/GigE-Cameras_AppNote_PIV-Min-Time-Between-Exposures.pdf retrieved 2024-04-05  # noqa: E501
    "Manta G-609": 47e-6,
    # cite: https://cdn.alliedvision.com/fileadmin/content/documents/products/cameras/Mako/techman/Mako_TechMan_en.pdf retrieved 2024-04-05  # noqa: E501
    "Mako G-040": 217e-6,
    "Mako G-125": 70e-6,
    "Mako G-234": 726e-6,
    "Mako G-507": 554e-6,
}


def get_camera_deadtime(model: str, override_deadtime: float | None = None) -> float:
    """Get the deadtime for a given camera model, or use an override value.

    :param model: The camera model string.
    :param override_deadtime:
        If provided, this value is returned instead of looking up the model.
    :returns: The deadtime in seconds.
    """
    if override_deadtime is not None:
        return override_deadtime
    elif model in _camera_deadtimes:
        return _camera_deadtimes[model]
    else:
        raise ValueError(f"Unknown camera model '{model}' for deadtime lookup.")



from ._detector import JungfrauDetector
from ._io import (
    AcquisitionType,
    DetectorStatus,
    GainMode,
    JungfrauDriverIO,
    JungfrauTriggerMode,
    PedestalMode,
)
from ._trigger_logic import JUNGFRAU_DEADTIME_S
from ._utils import (
    create_jungfrau_external_triggering_info,
    create_jungfrau_internal_triggering_info,
    create_jungfrau_pedestal_triggering_info,
)

__all__ = [
    "JungfrauDetector",
    "DetectorStatus",
    "create_jungfrau_external_triggering_info",
    "create_jungfrau_internal_triggering_info",
    "create_jungfrau_pedestal_triggering_info",
    "JungfrauDriverIO",
    "JungfrauTriggerMode",
    "AcquisitionType",
    "GainMode",
    "PedestalMode",
    "JUNGFRAU_DEADTIME_S",
]



import pytest
from pydantic import ValidationError

from ophyd_async.core import DetectorTrigger, TriggerInfo


@pytest.mark.parametrize(
    ["kwargs", "error_msg"],
    [
        (
            {
                "number_of_events": 1,
                "trigger": DetectorTrigger.EXTERNAL_LEVEL,
                "deadtime": 2,
                "livetime": 2,
                "exposure_timeout": "a",
            },
            "Input should be a valid number, unable to parse string as a number "
            "[type=float_parsing, input_value='a', input_type=str]",
        ),
        (
            {
                "number_of_events": 1,
                "trigger": "EXTERNAL_LEVEL",
                "deadtime": 2,
                "livetime": -1,
            },
            "Input should be greater than or equal to 0 "
            "[type=greater_than_equal, input_value=-1, input_type=int]",
        ),
        (
            {
                "number_of_events": 1,
                "trigger": DetectorTrigger.INTERNAL,
                "deadtime": 2,
                "livetime": 1,
                "exposure_timeout": -1,
            },
            "Input should be greater than 0 "
            "[type=greater_than, input_value=-1, input_type=int]",
        ),
        (
            {
                "number_of_events": 1,
                "trigger": "not_in_enum",
                "deadtime": 2,
                "livetime": 1,
                "exposure_timeout": None,
            },
            "Input should be 'INTERNAL', 'EXTERNAL_EDGE' or 'EXTERNAL_LEVEL' "
            "[type=enum, input_value='not_in_enum', input_type=str]",
        ),
        (
            {
                "number_of_events": -100,
            },
            "number_of_events\n  Input should be greater than or "
            "equal to 0 [type=greater_than_equal, input_value=-100, input_type=int]",
        ),
    ],
)
def test_malformed_trigger_info(kwargs, error_msg):
    with pytest.raises(ValidationError) as exc:
        TriggerInfo(**kwargs)
    assert error_msg in str(exc.value)



from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable, Mapping
from functools import cached_property
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from bluesky.protocols import Location, Reading, Subscribable
from event_model import DataKey

from ._protocol import AsyncLocatable, AsyncReadable
from ._signal_backend import SignalBackend, SignalDatatypeT, make_datakey, make_metadata
from ._utils import (
    Callback,
    ConfinedModel,
    T,
    error_if_none,
    gather_dict,
    merge_gathered_dicts,
)

RawT = TypeVar("RawT")
DerivedT = TypeVar("DerivedT")


class Transform(ConfinedModel, Generic[RawT, DerivedT]):
    """Baseclass for bidirectional transforms for Derived Signals.

    Subclass and add:
    - type hinted parameters that should be fetched from Signals
    - a raw_to_derived method that takes the elements of RawT and returns a DerivedT
    - a derived_to_raw method that takes the elements of DerivedT and returns a RawT

    :example:
    ```python
    class MyRaw(TypedDict):
        raw1: float
        raw2: float

    class MyDerived(TypedDict):
        derived1: float
        derived2: float

    class MyTransform(Transform):
        param1: float

        def raw_to_derived(self, *, raw1: float, raw2: float) -> MyDerived:
            derived1, derived2 = some_maths(self.param1, raw1, raw2)
            return MyDerived(derived1=derived1, derived2=derived2)

        def derived_to_raw(self, *, derived1: float, derived2: float) -> MyRaw:
            raw1, raw2 = some_inverse_maths(self.param1, derived1, derived2)
            return MyRaw(raw1=raw1, raw2=raw2)
    ```
    """

    if TYPE_CHECKING:
        # Guard with if type checking so they don't appear in pydantic argument list
        # Ideally they would be:
        #     def raw_to_derived(self, **kwargs: Unpack[RawT]) -> DerivedT: ...
        # but TypedDicts are not valid as generics
        # https://github.com/microsoft/pyright/discussions/7317
        raw_to_derived: Callable[..., DerivedT]
        derived_to_raw: Callable[..., RawT]


TransformT = TypeVar("TransformT", bound=Transform)


def validate_by_type(raw_devices: Mapping[str, Any], type_: type[T]) -> dict[str, T]:
    filtered_devices: dict[str, T] = {}
    for name, device in raw_devices.items():
        if not isinstance(device, type_):
            msg = f"{device} is not an instance of {type_}"
            raise TypeError(msg)
        filtered_devices[name] = device
    return filtered_devices


class SignalTransformer(Generic[TransformT]):
    def __init__(
        self,
        transform_cls: type[TransformT],
        set_derived: Callable[..., Awaitable[None]] | None,
        set_derived_takes_dict: bool,
        raw_devices,
        raw_constants,
        transform_devices,
        transform_constants,
    ):
        self._transform_cls = transform_cls
        self._set_derived = set_derived
        self._set_derived_takes_dict = set_derived_takes_dict

        self._transform_devices = transform_devices
        self._transform_constants = transform_constants
        self._raw_devices = raw_devices
        self._raw_constants = raw_constants

        self._derived_callbacks: dict[str, Callback[Reading]] = {}
        self._cached_readings: dict[str, Reading] | None = None

    @cached_property
    def raw_locatables(self) -> dict[str, AsyncLocatable]:
        return validate_by_type(self._raw_devices, AsyncLocatable)

    @cached_property
    def transform_readables(self) -> dict[str, AsyncReadable]:
        return validate_by_type(self._transform_devices, AsyncReadable)

    @cached_property
    def raw_and_transform_readables(self) -> dict[str, AsyncReadable]:
        return validate_by_type(
            self._raw_devices | self._transform_devices, AsyncReadable
        )

    @cached_property
    def raw_and_transform_subscribables(self) -> dict[str, Subscribable]:
        return validate_by_type(
            self._raw_devices | self._transform_devices, Subscribable
        )

    def _complete_cached_reading(self) -> dict[str, Reading] | None:
        if self._cached_readings and len(self._cached_readings) == len(
            self.raw_and_transform_subscribables
        ):
            return self._cached_readings
        return None

    def _make_transform_from_readings(
        self, transform_readings: dict[str, Reading]
    ) -> TransformT:
        # Make the transform using the values from the readings for those args
        transform_args = {
            k: transform_readings[sig.name]["value"]
            for k, sig in self.transform_readables.items()
        }
        return self._transform_cls(**(transform_args | self._transform_constants))

    def _make_derived_readings(
        self, raw_and_transform_readings: dict[str, Reading]
    ) -> dict[str, Reading]:
        # Calculate the latest timestamp and max severity from them
        timestamp = max(
            raw_and_transform_readings[device.name]["timestamp"]
            for device in self.raw_and_transform_subscribables.values()
        )
        alarm_severity = max(
            raw_and_transform_readings[device.name].get("alarm_severity", 0)
            for device in self.raw_and_transform_subscribables.values()
        )
        # Make the transform using the values from the readings for those args
        transform = self._make_transform_from_readings(raw_and_transform_readings)
        # Create the raw values from the rest then calculate the derived readings
        # using the transform
        # Extend dictionary with values of any Constants passed as arguments
        raw_values = {
            **{
                k: raw_and_transform_readings[sig.name]["value"]
                for k, sig in self._raw_devices.items()
            },
            **self._raw_constants,
        }

        derived_readings = {
            name: Reading(
                value=derived, timestamp=timestamp, alarm_severity=alarm_severity
            )
            for name, derived in transform.raw_to_derived(**raw_values).items()
        }
        return derived_readings

    async def get_transform(self) -> TransformT:
        if raw_and_transform_readings := self._complete_cached_reading():
            transform_readings = raw_and_transform_readings
        else:
            transform_readings = await merge_gathered_dicts(
                device.read() for device in self.transform_readables.values()
            )
        return self._make_transform_from_readings(transform_readings)

    async def get_derived_readings(self) -> dict[str, Reading]:
        if not (raw_and_transform_readings := self._complete_cached_reading()):
            raw_and_transform_readings = await merge_gathered_dicts(
                device.read() for device in self.raw_and_transform_readables.values()
            )
        return self._make_derived_readings(raw_and_transform_readings)

    async def get_derived_values(self) -> dict[str, Any]:
        derived_readings = await self.get_derived_readings()
        return {k: v["value"] for k, v in derived_readings.items()}

    def _update_cached_reading(self, value: dict[str, Reading]):
        _cached_readings = error_if_none(
            self._cached_readings,
            "Cannot update cached reading as it has not been initialised",
        )

        _cached_readings.update(value)
        if self._complete_cached_reading():
            # We've got a complete set of values, callback on them
            derived_readings = self._make_derived_readings(_cached_readings)
            for name, callback in self._derived_callbacks.items():
                callback(derived_readings[name])

    def set_callback(self, name: str, callback: Callback[Reading] | None) -> None:
        if callback is None:
            self._derived_callbacks.pop(name, None)
            if not self._derived_callbacks:
                # Remove the callbacks to all the raw devices
                for raw in self.raw_and_transform_subscribables.values():
                    raw.clear_sub(self._update_cached_reading)
                # and clear the cached readings that will now be stale
                self._cached_readings = None
        else:
            if name in self._derived_callbacks:
                msg = f"Callback already set for {name}"
                raise RuntimeError(msg)
            self._derived_callbacks[name] = callback
            if self._cached_readings is None:
                # Add the callbacks to all the raw devices, this will run the first
                # callback
                self._cached_readings = {}
                for raw in self.raw_and_transform_subscribables.values():
                    # can remove type: ignore when Subscribable protocol updated
                    raw.subscribe_reading(self._update_cached_reading)  # type: ignore
            elif self._complete_cached_reading():
                # Callback on the last complete set of readings
                derived_readings = self._make_derived_readings(self._cached_readings)
                callback(derived_readings[name])

    async def get_locations(self) -> dict[str, Location]:
        locations, transform = await asyncio.gather(
            gather_dict({k: sig.locate() for k, sig in self.raw_locatables.items()}),
            self.get_transform(),
        )
        raw_setpoints = {k: v["setpoint"] for k, v in locations.items()}
        raw_readbacks = {k: v["readback"] for k, v in locations.items()}
        derived_setpoints = transform.raw_to_derived(**raw_setpoints)
        derived_readbacks = transform.raw_to_derived(**raw_readbacks)
        return {
            name: Location(
                setpoint=derived_setpoints[name],
                readback=derived_readbacks[name],
            )
            for name in derived_setpoints
        }

    async def set_derived(self, name: str, value: Any):
        _set_derived = error_if_none(
            self._set_derived,
            "Cannot put as no set_derived method given",
        )
        if self._set_derived_takes_dict:
            # Need to get the other derived values and update the one that's changing
            derived = await self.get_locations()
            setpoints = {k: v["setpoint"] for k, v in derived.items()}
            setpoints[name] = value
            await _set_derived(setpoints)
        else:
            # Only one derived signal, so pass it directly
            await _set_derived(value)


class DerivedSignalBackend(SignalBackend[SignalDatatypeT]):
    def __init__(
        self,
        datatype: type[SignalDatatypeT],
        name: str,
        transformer: SignalTransformer,
        units: str | None = None,
        precision: int | None = None,
    ):
        self.name = name
        self.transformer = transformer
        # Add the extra static metadata to the dictionary
        self.metadata = make_metadata(datatype, units, precision)
        super().__init__(datatype)

    def source(self, name: str, read: bool) -> str:
        return f"derived://{name}"

    async def connect(self, timeout: float):
        # Assume that the underlying signals are already connected
        pass

    def set_value(self, value: SignalDatatypeT):
        msg = (
            "Cannot set the value of a derived signal, "
            "set the underlying raw signals instead"
        )
        raise RuntimeError(msg)

    async def put(self, value: SignalDatatypeT | None) -> None:
        value = error_if_none(
            value,
            "Must be given a value to put",
        )

        await self.transformer.set_derived(self.name, value)

    async def get_datakey(self, source: str) -> DataKey:
        return make_datakey(
            self.datatype or float, await self.get_value(), source, self.metadata
        )

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        readings = await self.transformer.get_derived_readings()
        return readings[self.name]

    async def get_value(self) -> SignalDatatypeT:
        derived = await self.transformer.get_derived_values()
        return derived[self.name]

    async def get_setpoint(self) -> SignalDatatypeT:
        # TODO: should be get_location
        locations = await self.transformer.get_locations()
        return locations[self.name]["setpoint"]

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        self.transformer.set_callback(self.name, callback)



from ._detector import EigerDetector
from ._io import EigerDetectorIO, EigerMonitorIO, EigerStreamIO, EigerTriggerMode

__all__ = [
    "EigerDetector",
    "EigerDetectorIO",
    "EigerMonitorIO",
    "EigerStreamIO",
    "EigerTriggerMode",
]



import asyncio
import os
from collections import defaultdict
from unittest.mock import ANY, Mock, call

import numpy as np
import pytest
from bluesky import plans as bp
from bluesky.protocols import Reading
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    LazyMock,
    get_mock,
    get_mock_put,
    init_devices,
    set_mock_value,
)
from ophyd_async.epics import demo
from ophyd_async.testing import (
    StatusWatcher,
    assert_configuration,
    assert_emitted,
    assert_reading,
    assert_value,
    wait_for_pending_wakeups,
)

# Can be removed once numpy >=2 is pinned.
scalar_int_dtype = (
    "<i4" if os.name == "nt" and np.version.version.startswith("1.") else "<i8"
)


@pytest.fixture
async def mock_motor():
    async with init_devices(mock=True):
        mock_motor = demo.DemoMotor("BLxxI-MO-TABLE-01:X:")
    set_mock_value(mock_motor.units, "mm")
    set_mock_value(mock_motor.precision, 3)
    set_mock_value(mock_motor.velocity, 1)
    yield mock_motor


@pytest.fixture
async def mock_point_detector():
    async with init_devices(mock=True):
        mock_point_detector = demo.DemoPointDetector("MOCK:DET:")
    yield mock_point_detector


async def test_motor_stopped(mock_motor: demo.DemoMotor):
    # Check it hasn't already been called
    stop_mock = get_mock_put(mock_motor.stop_)
    stop_mock.assert_not_called()
    # Call stop and check it's called with the default value
    await mock_motor.stop()
    stop_mock.assert_called_once_with(None)
    # We can also track all the mock puts that have happened on the device
    parent_mock = get_mock(mock_motor)
    await mock_motor.velocity.set(15)
    assert parent_mock.mock_calls == [
        call.stop_.put(None),
        call.velocity.put(15),
    ]


async def test_motor_moving_well(mock_motor: demo.DemoMotor) -> None:
    # Start it moving
    s = mock_motor.set(0.55)
    # Watch for updates, and make sure the first update is the current position
    watcher = StatusWatcher(s)
    await watcher.wait_for_call(
        name="mock_motor",
        current=0.0,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.0, abs=0.08),
    )
    await assert_value(mock_motor.setpoint, 0.55)
    assert not s.done
    # Wait a bit and give it an update, checking that the watcher is called with it
    await asyncio.sleep(0.1)
    set_mock_value(mock_motor.readback, 0.1)
    await watcher.wait_for_call(
        name="mock_motor",
        current=0.1,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.1, abs=0.08),
    )
    # Make it almost get there and check that it completes
    set_mock_value(mock_motor.readback, 0.5499999)
    await wait_for_pending_wakeups()
    assert s.done
    assert s.success


async def test_retrieve_mock_and_assert(mock_motor: demo.DemoMotor):
    motor_setpoint_mock = get_mock_put(mock_motor.setpoint)
    await mock_motor.setpoint.set(10)
    motor_setpoint_mock.assert_called_once_with(10)

    # Assert that velocity is set before move
    motor_velocity_mock = get_mock_put(mock_motor.velocity)

    parent_mock = Mock()
    parent_mock.attach_mock(motor_setpoint_mock, "setpoint")
    parent_mock.attach_mock(motor_velocity_mock, "velocity")

    await mock_motor.velocity.set(100)
    await mock_motor.setpoint.set(67)
    assert parent_mock.mock_calls == [
        call.velocity(100),
        call.setpoint(67),
    ]


async def test_mocks_in_device_share_parent():
    lm = LazyMock()
    mock_motor = demo.DemoMotor("BLxxI-MO-TABLE-01:Y:")
    await mock_motor.connect(mock=lm)
    mock = lm()

    assert get_mock(mock_motor) is mock
    assert get_mock(mock_motor.setpoint) is mock.setpoint
    assert get_mock_put(mock_motor.setpoint) is mock.setpoint.put
    await mock_motor.setpoint.set(10)
    get_mock_put(mock_motor.setpoint).assert_called_once_with(10)

    await mock_motor.velocity.set(100)
    await mock_motor.setpoint.set(67)

    mock.reset_mock()
    await mock_motor.velocity.set(100)
    await mock_motor.setpoint.set(67)
    assert mock.mock_calls == [
        call.velocity.put(100),
        call.setpoint.put(67),
    ]


async def test_read_motor(mock_motor: demo.DemoMotor):
    await mock_motor.stage()
    await assert_reading(
        mock_motor,
        {"mock_motor": {"value": 0.0, "timestamp": ANY, "alarm_severity": 0}},
    )
    await assert_configuration(
        mock_motor,
        {
            "mock_motor-units": {
                "value": "mm",
                "timestamp": ANY,
                "alarm_severity": 0,
            },
            "mock_motor-velocity": {
                "value": 1.0,
                "timestamp": ANY,
                "alarm_severity": 0,
            },
        },
    )
    # Check that changing the readback value changes the reading
    set_mock_value(mock_motor.readback, 0.5)
    await assert_value(mock_motor.readback, 0.5)
    await assert_reading(
        mock_motor,
        {"mock_motor": {"value": 0.5, "timestamp": ANY, "alarm_severity": 0}},
    )
    # Check we can still read when not staged
    await mock_motor.unstage()
    set_mock_value(mock_motor.readback, 0.1)
    await assert_reading(
        mock_motor,
        {"mock_motor": {"value": 0.1, "timestamp": ANY, "alarm_severity": 0}},
    )


async def test_set_velocity(mock_motor: demo.DemoMotor) -> None:
    v = mock_motor.velocity
    q: asyncio.Queue[dict[str, Reading]] = asyncio.Queue()
    v.subscribe_reading(q.put_nowait)
    assert (await q.get())["mock_motor-velocity"]["value"] == 1.0
    await v.set(2.0)
    assert (await q.get())["mock_motor-velocity"]["value"] == 2.0
    v.clear_sub(q.put_nowait)
    await v.set(3.0)
    assert (await v.read())["mock_motor-velocity"]["value"] == 3.0
    assert q.empty()


async def test_zero_velocity(mock_motor: demo.DemoMotor) -> None:
    # v = sim_motor.velocity
    await mock_motor.velocity.set(0)
    with pytest.raises(ZeroDivisionError):
        await mock_motor.set(3.14)


async def test_read_point_detector(mock_point_detector: demo.DemoPointDetector):
    channel = mock_point_detector.channel[1]
    assert (await channel.read())["mock_point_detector-channel-1-value"]["value"] == 0
    assert (await channel.read_configuration())["mock_point_detector-channel-1-mode"][
        "value"
    ] == demo.EnergyMode.LOW
    desc = (await channel.describe_configuration())[
        "mock_point_detector-channel-1-mode"
    ]
    assert desc["dtype"] == "string"
    assert desc["choices"] == ["Low Energy", "High Energy"]
    set_mock_value(channel.mode, demo.EnergyMode.HIGH)
    assert (await channel.read_configuration())["mock_point_detector-channel-1-mode"][
        "value"
    ] == demo.EnergyMode.HIGH


async def test_point_detector_in_plan(
    RE: RunEngine, mock_point_detector: demo.DemoPointDetector
):
    # Subscribe to new documents produce, putting them in a dict by type
    docs = defaultdict(list)
    RE.subscribe(lambda name, doc: docs[name].append(doc))
    # Set the channel values to a known value
    for i, channel in mock_point_detector.channel.items():
        set_mock_value(channel.value, 100 + i)
    # Run the plan and assert the right docs are produced
    RE(bp.count([mock_point_detector], num=2))
    assert_emitted(docs, start=1, descriptor=1, event=2, stop=1)
    assert docs["event"][1]["data"] == {
        "mock_point_detector-channel-1-value": 101,
        "mock_point_detector-channel-2-value": 102,
        "mock_point_detector-channel-3-value": 103,
    }


async def test_assembly_renaming() -> None:
    thing = demo.DemoStage("PRE")
    await thing.connect(mock=True)
    assert thing.x.name == ""
    assert thing.x.velocity.name == ""
    assert thing.x.stop_.name == ""
    await thing.x.velocity.set(456)
    assert await thing.x.velocity.get_value() == 456
    thing.set_name("foo")
    assert thing.x.name == "foo-x"
    assert thing.x.velocity.name == "foo-x-velocity"
    assert thing.x.stop_.name == "foo-x-stop_"


async def test_point_detector_read_and_describe(
    mock_point_detector: demo.DemoPointDetector,
):
    set_mock_value(mock_point_detector.channel[1].value, 1)
    set_mock_value(mock_point_detector.channel[2].value, 5)
    set_mock_value(mock_point_detector.channel[3].value, 10)

    await mock_point_detector.stage()
    description = await mock_point_detector.describe()

    await mock_point_detector.unstage()
    await assert_reading(
        mock_point_detector,
        {
            "mock_point_detector-channel-1-value": {
                "value": 1,
                "timestamp": ANY,
                "alarm_severity": 0,
            },
            "mock_point_detector-channel-2-value": {
                "value": 5,
                "timestamp": ANY,
                "alarm_severity": 0,
            },
            "mock_point_detector-channel-3-value": {
                "value": 10,
                "timestamp": ANY,
                "alarm_severity": 0,
            },
        },
    )
    assert description == {
        "mock_point_detector-channel-1-value": {
            "dtype": "integer",
            "dtype_numpy": scalar_int_dtype,
            "shape": [],
            "source": "mock+ca://MOCK:DET:1:Value",
        },
        "mock_point_detector-channel-2-value": {
            "dtype": "integer",
            "dtype_numpy": scalar_int_dtype,
            "shape": [],
            "source": "mock+ca://MOCK:DET:2:Value",
        },
        "mock_point_detector-channel-3-value": {
            "dtype": "integer",
            "dtype_numpy": scalar_int_dtype,
            "shape": [],
            "source": "mock+ca://MOCK:DET:3:Value",
        },
    }



import re
import uuid
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from datetime import date
from pathlib import Path, PurePath, PureWindowsPath
from typing import Protocol
from urllib.parse import urlunparse


@dataclass
class PathInfo:
    """Information about where and how to write a file.

    :param directory_path: Directory into which files should be written
    :param filename: Base filename to use generated by FilenameProvider, w/o extension
    :param create_dir_depth: Optional depth of directories to create if they do not
                             exist
    :param directory_uri: Optional URI to use for reading back resources. If not set,
                          it will be generated from the directory path.
    """

    directory_path: PurePath
    filename: str
    create_dir_depth: int = 0
    directory_uri: str | None = None

    def __post_init__(self):
        if not self.directory_path.is_absolute():
            raise ValueError(
                f"directory_path must be an absolute path, got {self.directory_path}"
            )

        # If directory uri is not set, set it using the directory path.
        if self.directory_uri is None:
            self.directory_uri = urlunparse(
                (
                    "file",
                    "localhost",
                    f"{self.directory_path.as_posix()}/",
                    "",
                    "",
                    None,
                )
            )
        elif not self.directory_uri.endswith("/"):
            # Ensure the directory URI ends with a slash.
            self.directory_uri += "/"


class FilenameProvider(Protocol):
    """Base class for callable classes providing filenames."""

    @abstractmethod
    def __call__(self, datakey_name: str | None = None) -> str:
        """Get a filename to use for output data, w/o extension."""


class PathProvider(Protocol):
    """Abstract class that tells a detector where to write its data."""

    @abstractmethod
    def __call__(self, datakey_name: str | None = None) -> PathInfo:
        """Get the current directory to write files into."""


class StaticFilenameProvider(FilenameProvider):
    """Provides a constant filename on every call."""

    def __init__(self, filename: str):
        self._static_filename = filename

    def __call__(self, datakey_name: str | None = None) -> str:
        return self._static_filename


class UUIDFilenameProvider(FilenameProvider):
    """Files will have a UUID as a filename."""

    def __init__(
        self,
        uuid_call_func: Callable = uuid.uuid4,
        uuid_call_args: list | None = None,
    ):
        self._uuid_call_func = uuid_call_func
        self._uuid_call_args = uuid_call_args or []

    def __call__(self, datakey_name: str | None = None) -> str:
        if (
            self._uuid_call_func in [uuid.uuid3, uuid.uuid5]
            and len(self._uuid_call_args) < 2
        ):
            raise ValueError(
                f"To use {self._uuid_call_func} to generate UUID filenames,"
                " UUID namespace and name must be passed as args!"
            )

        uuid_str = self._uuid_call_func(*self._uuid_call_args)
        return f"{uuid_str}"


class AutoIncrementFilenameProvider(FilenameProvider):
    """Provides a new numerically incremented filename on each call."""

    def __init__(
        self,
        base_filename: str = "",
        max_digits: int = 5,
        starting_value: int = 0,
        increment: int = 1,
        inc_delimeter: str = "_",
    ):
        self._base_filename = base_filename
        self._max_digits = max_digits
        self._current_value = starting_value
        self._increment = increment
        self._inc_delimeter = inc_delimeter

    def __call__(self, datakey_name: str | None = None) -> str:
        if len(str(self._current_value)) > self._max_digits:
            raise ValueError(
                f"Auto incrementing filename counter "
                f"exceeded maximum of {self._max_digits} digits!"
            )

        padded_counter = f"{self._current_value:0{self._max_digits}}"

        filename = f"{self._base_filename}{self._inc_delimeter}{padded_counter}"

        self._current_value += self._increment
        return filename


class StaticPathProvider(PathProvider):
    """All files will be within a static directory."""

    def __init__(
        self,
        filename_provider: FilenameProvider,
        directory_path: PurePath,
        directory_uri: str | None = None,
        create_dir_depth: int = 0,
    ) -> None:
        self._filename_provider = filename_provider
        self._directory_path = directory_path
        self._directory_uri = directory_uri
        self._create_dir_depth = create_dir_depth

    def __call__(self, datakey_name: str | None = None) -> PathInfo:
        filename = self._filename_provider(datakey_name)

        return PathInfo(
            directory_path=self._directory_path,
            directory_uri=self._directory_uri,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )


class AutoMaxIncrementingPathProvider(PathProvider):
    """Increment directory name on each call by checking existing directories.

    Looks through directories in a specified base path to increment directory name.
    PathInfo gives path like base_path/0001_dirname/dirname, or
    base_path/yyyy-mm-dd/0001_dirname/dirname if dated is true. Here,
    the '0001' is the value which gets incremented if the file exists already.

    It's recommended for the base path provider to be non-incrementing.

    Args:
    base_path_provider: Path to create directories inside of. Note that the filename of
    this provider is used as the top level directory and the filename
    max_digits: Number of digits to pad onto the parent directory.
    starting_value: Number to start incrementing from.
    dated: Whether to create an extra directory to specify the day.

    """

    def __init__(
        self,
        base_path_provider: PathProvider,
        max_digits: int = 4,
        starting_value: int = 0,
        dated: bool = False,
    ):
        self._base_path_provider = base_path_provider
        self._max_digits = max_digits
        self._next_value = starting_value
        self._dated = dated

    def _get_highest_number_from(self, path: Path) -> int:
        # Look through directories in path which end in "_{number} and get highest
        # number"
        highest_number = 0
        candidates = [
            x for x in path.iterdir() if x.is_dir() and re.match(r"^\d+_", x.name)
        ]
        if candidates:
            highest_number = max(
                int(x.name.split("_", maxsplit=1)[0]) for x in candidates
            )
        else:
            highest_number = self._next_value
        return highest_number

    def __call__(self, datakey_name: str | None = None) -> PathInfo:
        base_path_info = self._base_path_provider.__call__()
        base_path_dir = Path(base_path_info.directory_path)
        if self._dated:
            # Make sure we are the max ID of any other days to keep numbering
            # consistent.
            cands = [
                self._get_highest_number_from(x)
                for x in base_path_dir.iterdir()
                if re.match(r"^\d\d\d\d-\d\d-\d\d$", x.name)
            ]
            if cands:
                self._next_value = max(max(cands) + 1, self._next_value)

            path = base_path_dir / date.today().strftime("%Y-%m-%d")
        else:
            path = base_path_dir

        val_to_use = self._next_value

        # Get the highest number using files in path, or use
        # stored next value if no files are found.
        if path.exists():
            highest_number = self._get_highest_number_from(path)
            val_to_use = (
                highest_number + 1
                if not highest_number == self._next_value
                else highest_number
            )

        self._next_value = val_to_use + 1

        filename = base_path_info.filename

        padded_counter = f"{val_to_use:0{self._max_digits}}"
        full_path = (
            path / f"{padded_counter}_{filename.strip('_')}" / filename.rstrip("_")
        )
        return PathInfo(
            directory_path=full_path.parent,
            directory_uri=None,
            filename=full_path.name,
            create_dir_depth=0,
        )


class AutoIncrementingPathProvider(PathProvider):
    """Provides a new numerically incremented path on each call."""

    def __init__(
        self,
        filename_provider: FilenameProvider,
        base_directory_path: PurePath,
        base_directory_uri: str | None = None,
        create_dir_depth: int = 0,
        max_digits: int = 5,
        starting_value: int = 0,
        num_calls_per_inc: int = 1,
        increment: int = 1,
        inc_delimeter: str = "_",
        base_name: str | None = None,
    ) -> None:
        self._filename_provider = filename_provider
        self._base_directory_path = base_directory_path
        self._base_directory_uri = base_directory_uri
        if (
            self._base_directory_uri is not None
            and not self._base_directory_uri.endswith("/")
        ):
            self._base_directory_uri += "/"
        self._create_dir_depth = create_dir_depth
        self._base_name = base_name
        self._starting_value = starting_value
        self._current_value = starting_value
        self._num_calls_per_inc = num_calls_per_inc
        self._inc_counter = 0
        self._max_digits = max_digits
        self._increment = increment
        self._inc_delimeter = inc_delimeter

    def __call__(self, datakey_name: str | None = None) -> PathInfo:
        filename = self._filename_provider(datakey_name)

        padded_counter = f"{self._current_value:0{self._max_digits}}"

        auto_inc_dir_name = str(padded_counter)
        if self._base_name is not None:
            auto_inc_dir_name = (
                f"{self._base_name}{self._inc_delimeter}{padded_counter}"
            )
        elif datakey_name is not None:
            auto_inc_dir_name = f"{datakey_name}{self._inc_delimeter}{padded_counter}"

        self._inc_counter += 1
        if self._inc_counter == self._num_calls_per_inc:
            self._inc_counter = 0
            self._current_value += self._increment

        directory_uri = None
        if self._base_directory_uri is not None:
            directory_uri = f"{self._base_directory_uri}{auto_inc_dir_name}"

        return PathInfo(
            directory_path=self._base_directory_path / auto_inc_dir_name,
            directory_uri=directory_uri,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )


class YMDPathProvider(PathProvider):
    """Provides a path with the date included in the directory name."""

    def __init__(
        self,
        filename_provider: FilenameProvider,
        base_directory_path: PurePath,
        base_directory_uri: str | None = None,
        create_dir_depth: int = -3,  # Default to -3 to create YMD dirs
        datakey_name_as_base_dir: bool = False,
    ) -> None:
        self._filename_provider = filename_provider
        self._base_directory_path = base_directory_path
        self._base_directory_uri = base_directory_uri
        if (
            self._base_directory_uri is not None
            and not self._base_directory_uri.endswith("/")
        ):
            self._base_directory_uri += "/"
        self._create_dir_depth = create_dir_depth
        self._datakey_name_as_base_dir = datakey_name_as_base_dir

    def __call__(self, datakey_name: str | None = None) -> PathInfo:
        path_type = type(self._base_directory_path)
        if path_type == PureWindowsPath:
            sep = "\\"
        else:
            sep = "/"

        current_date = date.today().strftime(f"%Y{sep}%m{sep}%d")
        if datakey_name is None:
            ymd_dir_path = current_date
        elif self._datakey_name_as_base_dir:
            ymd_dir_path = path_type(
                current_date,
                datakey_name,
            )
        else:
            ymd_dir_path = path_type(
                datakey_name,
                current_date,
            )

        filename = self._filename_provider(datakey_name)

        directory_uri = None
        if self._base_directory_uri is not None:
            directory_uri = f"{self._base_directory_uri}{ymd_dir_path}"

        return PathInfo(
            directory_path=self._base_directory_path / ymd_dir_path,
            directory_uri=directory_uri,
            filename=filename,
            create_dir_depth=self._create_dir_depth,
        )



from collections.abc import Sequence
from pathlib import Path
from typing import Annotated as A

import numpy as np

from ophyd_async.core import Array1D, SignalR, SignalRW, StrictEnum, SubsetEnum, Table
from ophyd_async.epics.core import EpicsDevice, PvSuffix

from ._utils import TestingIOC, generate_random_pv_prefix

CA_PVA_RECORDS = Path(__file__).parent / "test_records.db"
PVA_RECORDS = Path(__file__).parent / "test_records_pva.db"


class EpicsTestEnum(StrictEnum):
    """For testing strict enum values in test IOCs."""

    A = "Aaa"
    B = "Bbb"
    C = "Ccc"


class EpicsTestSubsetEnum(SubsetEnum):
    """For testing subset enum values in test IOCs."""

    A = "Aaa"
    B = "Bbb"


class EpicsTestTable(Table):
    a_bool: Array1D[np.bool_]
    a_int: Array1D[np.int32]
    a_float: Array1D[np.float64]
    a_str: Sequence[str]
    a_enum: Sequence[EpicsTestEnum]


class EpicsTestCaDevice(EpicsDevice):
    """Device for use in a channel access test IOC."""

    a_int: A[SignalRW[int], PvSuffix("int")]
    """A thing"""
    a_float: A[SignalRW[float], PvSuffix("float")]
    float_prec_0: A[SignalRW[int], PvSuffix("float_prec_0")]
    a_str: A[SignalRW[str], PvSuffix("str")]
    longstr: A[SignalRW[str], PvSuffix("longstr")]
    longstr2: A[SignalRW[str], PvSuffix("longstr2.VAL$")]
    a_bool: A[SignalRW[bool], PvSuffix("bool")]
    slowseq: A[SignalRW[int], PvSuffix("slowseq")]
    enum: A[SignalRW[EpicsTestEnum], PvSuffix("enum")]
    enum2: A[SignalRW[EpicsTestEnum], PvSuffix("enum2")]
    subset_enum: A[SignalRW[EpicsTestSubsetEnum], PvSuffix("subset_enum")]
    enum_str_fallback: A[SignalRW[str], PvSuffix("enum_str_fallback")]
    bool_unnamed: A[SignalRW[bool], PvSuffix("bool_unnamed")]
    partialint: A[SignalRW[int], PvSuffix("partialint")]
    lessint: A[SignalRW[int], PvSuffix("lessint")]
    uint8a: A[SignalRW[Array1D[np.uint8]], PvSuffix("uint8a")]
    int16a: A[SignalRW[Array1D[np.int16]], PvSuffix("int16a")]
    int32a: A[SignalRW[Array1D[np.int32]], PvSuffix("int32a")]
    float32a: A[SignalRW[Array1D[np.float32]], PvSuffix("float32a")]
    float64a: A[SignalRW[Array1D[np.float64]], PvSuffix("float64a")]
    stra: A[SignalRW[Sequence[str]], PvSuffix("stra")]


class EpicsTestPvaDevice(EpicsTestCaDevice):
    """Device for use in a pv access test IOC."""

    # pva can support all signal types that ca can
    int8a: A[SignalRW[Array1D[np.int8]], PvSuffix("int8a")]
    uint16a: A[SignalRW[Array1D[np.uint16]], PvSuffix("uint16a")]
    uint32a: A[SignalRW[Array1D[np.uint32]], PvSuffix("uint32a")]
    int64a: A[SignalRW[Array1D[np.int64]], PvSuffix("int64a")]
    uint64a: A[SignalRW[Array1D[np.uint64]], PvSuffix("uint64a")]
    table: A[SignalRW[EpicsTestTable], PvSuffix("table")]
    ntndarray: A[SignalR[np.ndarray], PvSuffix("ntndarray")]


class EpicsTestIocAndDevices:
    """Test IOC with ca and pva devices."""

    def __init__(self):
        self.prefix = generate_random_pv_prefix()
        self.ioc = TestingIOC()
        # Create supporting records and ExampleCaDevice
        ca_prefix = f"{self.prefix}ca:"
        self.ioc.add_database(CA_PVA_RECORDS, device=ca_prefix)
        self.ca_device = EpicsTestCaDevice(f"ca://{ca_prefix}")
        # Create supporting records and ExamplePvaDevice
        pva_prefix = f"{self.prefix}pva:"
        self.ioc.add_database(CA_PVA_RECORDS, device=pva_prefix)
        self.ioc.add_database(PVA_RECORDS, device=pva_prefix)
        self.pva_device = EpicsTestPvaDevice(f"pva://{pva_prefix}")

    def get_device(self, protocol: str) -> EpicsTestCaDevice | EpicsTestPvaDevice:
        return getattr(self, f"{protocol}_device")

    def get_signal(self, protocol: str, name: str) -> SignalRW:
        return getattr(self.get_device(protocol), name)

    def get_pv(self, protocol: str, name: str) -> str:
        return f"{protocol}://{self.prefix}{protocol}:{name}"



import asyncio
import os
from collections import defaultdict

import bluesky.plan_stubs as bps
import bluesky.plans as bp
import bluesky.preprocessors as bpp
import h5py
import numpy as np
import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import StaticFilenameProvider, StaticPathProvider, TriggerInfo
from ophyd_async.sim import SimBlobDetector
from ophyd_async.testing import assert_emitted


@pytest.fixture
def blob_detector(tmp_path) -> SimBlobDetector:
    path_provider = StaticPathProvider(StaticFilenameProvider("file"), tmp_path)
    return SimBlobDetector(path_provider, name="det")


@pytest.mark.parametrize("x_position,det_sum", [(0.0, 277544), (1.0, 506344)])
async def test_sim_blob_detector_count(
    RE: RunEngine,
    tmp_path,
    x_position: float,
    det_sum: int,
    blob_detector: SimBlobDetector,
):
    docs = defaultdict(list)
    RE.subscribe(lambda name, doc: docs[name].append(doc))
    blob_detector.pattern_generator.set_x(x_position)

    RE(bp.count([blob_detector], num=2))
    assert_emitted(
        docs, start=1, descriptor=1, stream_resource=2, stream_datum=4, event=2, stop=1
    )
    uri = docs["stream_resource"][0]["uri"]
    path = uri.split("://localhost")[-1]
    if os.name == "nt":
        path = path.lstrip("/")
    # Check data looks right
    assert path == (tmp_path / "file.h5").as_posix()
    h5file = h5py.File(path)
    assert list(h5file["/entry"]) == ["data", "sum"]
    assert list(h5file["/entry/sum"]) == [det_sum, det_sum]
    assert np.sum(h5file["/entry/data/data"][0]) == det_sum
    # Check descriptor looks right
    assert docs["descriptor"][0]["data_keys"] == {
        "det": {
            "source": uri,
            "shape": [1, 240, 320],
            "dtype": "array",
            "dtype_numpy": "|u1",
            "object_name": "det",
            "external": "STREAM:",
        },
        "det-sum": {
            "source": uri,
            "shape": [1],
            "dtype": "number",
            "dtype_numpy": "<i8",
            "object_name": "det",
            "external": "STREAM:",
        },
    }


async def test_sim_blob_detector_fly(RE: RunEngine, blob_detector: SimBlobDetector):
    @bpp.stage_decorator([blob_detector])
    @bpp.run_decorator()
    def fly_plan():
        yield from bps.prepare(
            blob_detector, TriggerInfo(number_of_events=7), wait=True
        )
        yield from bps.declare_stream(blob_detector, name="primary")
        yield from bps.kickoff(blob_detector, wait=True)
        yield from bps.collect_while_completing(
            flyers=[blob_detector], dets=[blob_detector], flush_period=0.1
        )

    docs = defaultdict(list)
    RE.subscribe(lambda name, doc: docs[name].append(doc))

    values = iter(np.linspace(1, 2, 7))

    async def set_next_x(t: float):
        x = next(values)
        blob_detector.pattern_generator.set_x(x)
        if x == pytest.approx(1.5):
            # pause here so we emit the data in 2 stream_datums
            await asyncio.sleep(0.2)

    blob_detector.pattern_generator.sleep = set_next_x

    RE(fly_plan())
    assert_emitted(
        docs, start=1, descriptor=1, stream_resource=2, stream_datum=4, stop=1
    )
    # Check we get the right numbers of frames in each stream_datum
    assert (
        docs["stream_datum"][0]["indices"]
        == docs["stream_datum"][1]["indices"]
        == {
            "start": 0,
            "stop": 3,
        }
    )
    assert (
        docs["stream_datum"][2]["indices"]
        == docs["stream_datum"][3]["indices"]
        == {
            "start": 3,
            "stop": 7,
        }
    )
    # Check we get the right data
    path = docs["stream_resource"][0]["uri"].split("://localhost")[-1]
    if os.name == "nt":
        path = path.lstrip("/")
    h5file = h5py.File(path)
    assert list(h5file["/entry/sum"]) == [
        506344,
        527596,
        542140,
        549008,
        548516,
        540424,
        524808,
    ]



import asyncio
import sys
from unittest.mock import patch

import pytest
from bluesky.plans import spiral_square
from bluesky.run_engine import RunEngine

from ophyd_async.core import FlyMotorInfo
from ophyd_async.sim import SimMotor
from ophyd_async.testing import StatusWatcher


async def test_move_sim_in_plan():
    RE = RunEngine()
    m1 = SimMotor("M1")
    m2 = SimMotor("M2")

    my_plan = spiral_square([], m1, m2, 0, 0, 4, 4, 10, 10)

    RE(my_plan)

    assert await m1.user_readback.get_value() == -2
    assert await m2.user_readback.get_value() == -2


@pytest.fixture
def m1() -> SimMotor:
    return SimMotor("M1", instant=False)


@pytest.mark.xfail(reason="Flaky test")
@pytest.mark.skipif("win" in sys.platform, reason="windows CI runners too weedy")
@pytest.mark.parametrize(
    "setpoint,expected",
    [
        (-0.19, [0.0, -0.05, -0.1495, -0.19]),
        (0.26, [0.0, 0.05, 0.15, 0.242, 0.26]),
        (0.005, [0.0, 0.005]),
        (-0.025, [0.0, -0.025]),
    ],
)
async def test_move_profiles(setpoint, expected, m1: SimMotor):
    await m1.acceleration_time.set(0.1)
    status = m1.set(setpoint)
    watcher = StatusWatcher(status)
    for i, v in enumerate(expected):
        await watcher.wait_for_call(
            current=pytest.approx(v),
            initial=0.0,
            name="M1",
            target=setpoint,
            time_elapsed=pytest.approx(i * 0.1, abs=0.1),
            unit="mm",
        )
    await status
    watcher.mock.assert_not_called()
    assert await m1.user_readback.get_value() == setpoint


async def test_short_move_is_exactly_move_time(m1: SimMotor):
    with patch("asyncio.sleep") as mock_sleep:
        await m1.set(0.0032)
    mock_sleep.assert_called_once_with(pytest.approx(0.08, abs=0.02))


@pytest.mark.timeout(3)
async def test_stop(m1: SimMotor):
    # this move should take 10 seconds but we will stop it after 0.5
    move_status = m1.set(10)
    await asyncio.sleep(0.5)
    await m1.stop(success=False)
    new_pos = await m1.user_readback.get_value()
    assert new_pos < 10
    assert new_pos >= 0.1

    assert not move_status.success
    with pytest.raises(RuntimeError, match="Motor was stopped"):
        await move_status


@pytest.mark.skipif("win" in sys.platform, reason="windows CI runners too weedy")
async def test_fly(m1: SimMotor):
    await m1.acceleration_time.set(0.1)
    info = FlyMotorInfo(start_position=0, end_position=1, time_for_move=0.2)
    fly_start, fly_end, velocity = -0.25, 1.25, 5
    await m1.prepare(info)
    assert await m1.user_readback.get_value() == fly_start
    assert await m1.velocity.get_value() == velocity
    await m1.kickoff()
    status = m1.complete()
    watcher = StatusWatcher(status)
    for i, v in enumerate([-0.25, 0, 0.5, 1.0, 1.25]):
        await watcher.wait_for_call(
            current=pytest.approx(v),
            initial=fly_start,
            name="M1",
            target=fly_end,
            time_elapsed=pytest.approx(i * 0.1, abs=0.1),
            unit="mm",
        )
    await status
    watcher.mock.assert_not_called()
    assert await m1.user_readback.get_value() == fly_end


async def test_sim_motor_can_be_set_to_its_current_position(m1: SimMotor):
    await m1.set(0)
